/***************************************************************
 * (C) COPYRIGHT SerComm Corporation All Rights Reserved, 2006.
 *
 *      Use of this software is restricted to the terms and
 *      conditions of SerComm's software license agreement.
 *
 *                        www.sercomm.com
 ****************************************************************/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/poll.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/atomic.h>
#include <asm/dma.h>
#include <asm/irq.h>
#include <linux/sound.h>
#include <linux/soundcard.h>
#include <linux/major.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/arch-pl1029/pl_symbol_alarm.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <linux/i2c-proc.h>
#include <linux/pm.h>
#include <linux/delay.h>
#include "tlv320aic23b.h"

//#define DEBUG
//#ifdef DEBUG
#if 0
#define	AUDIO_DBG(fmt, args...)	{		\
	printk("%s(%d): ", __FILE__, __LINE__);	\
	printk(fmt, ##args);					\
}
#define	AUDIO_LINE	printk("(%s:%d) @ %s\n", __FUNCTION__, __LINE__, __FILE__)
#else
#define	AUDIO_DBG(fmt, args...)
#define	AUDIO_LINE
#endif

#define _AIC23HW_BUG_
/******************************************************************************
 * Prolific AC'97/I2S/SPDIF Module Regsiter Definition
 ******************************************************************************/

#define PL_AUDIO_BASE_IO_PORT       0xd9480000
#define PL_AUDIO_OB                 (PL_AUDIO_BASE_IO_PORT + 0x00000)
#define PL_AUDIO_IB                 (PL_AUDIO_BASE_IO_PORT + 0x10000)
#define PL_AUDIO_CMD                (PL_AUDIO_BASE_IO_PORT + 0x20000)
#define PL_AUDIO_STATUS             (PL_AUDIO_BASE_IO_PORT + 0x30000)

/* CMD Register Set */
#define PL_AUDIO_CMD_COMMAND_0      (PL_AUDIO_CMD + 0x00)
#define PL_AUDIO_CMD_COMMAND_1      (PL_AUDIO_CMD + 0x01)
#define PL_AUDIO_CMD_COMMAND_2      (PL_AUDIO_CMD + 0x02)
#define PL_AUDIO_CMD_COMMAND_3      (PL_AUDIO_CMD + 0x03)
#define PL_AUDIO_CMD_FUNCTION_0     (PL_AUDIO_CMD + 0x04 + 0x00)
#define PL_AUDIO_CMD_FUNCTION_1     (PL_AUDIO_CMD + 0x04 + 0x01)
#define PL_AUDIO_CMD_FUNCTION_2     (PL_AUDIO_CMD + 0x04 + 0x02)
#define PL_AUDIO_CMD_FUNCTION_3     (PL_AUDIO_CMD + 0x04 + 0x03)
#define PL_AUDIO_CMD_CONFIG_0   	 (PL_AUDIO_CMD + 0x08 + 0x00)
#define PL_AUDIO_CMD_CONFIG_1   (PL_AUDIO_CMD + 0x08 + 0x01)
#define PL_AUDIO_CMD_CONFIG_2   (PL_AUDIO_CMD + 0x08 + 0x02)
#define PL_AUDIO_CMD_CONFIG_3   (PL_AUDIO_CMD + 0x08 + 0x03)
#define PL_AUDIO_CMD_AC97ADDR       (PL_AUDIO_CMD + 0x0c)
#define PL_AUDIO_CMD_AC97WDATA      (PL_AUDIO_CMD + 0x10)
#define PL_AUDIO_CMD_AC97STATUS     (PL_AUDIO_CMD + 0x14)
#define PL_AUDIO_CMD_AC97RDATA      (PL_AUDIO_CMD + 0x18)
#define PL_AUDIO_CMD_CLOCK      (PL_AUDIO_CMD + 0x1c)
#define PL_AUDIO_CMD_GPIO_C_ENB     (PL_AUDIO_CMD + 0x20 + 0x00)
#define PL_AUDIO_CMD_GPIO_C_DO      (PL_AUDIO_CMD + 0x20 + 0x01)
#define PL_AUDIO_CMD_GPIO_C_OE      (PL_AUDIO_CMD + 0x24)
#define PL_AUDIO_CMD_GPIO_C_DRV_E2  (PL_AUDIO_CMD + 0x24 + 0x01)
#define PL_AUDIO_CMD_GPIO_C_DRV_E4  (PL_AUDIO_CMD + 0x24 + 0x02)
#define PL_AUDIO_CMD_GPIO_C_DRV_E8  (PL_AUDIO_CMD + 0x24 + 0x03)
#define PL_AUDIO_CMD_GPIOPULL       (PL_AUDIO_CMD + 0x28)
#define PL_AUDIO_CMD_GPIOIN         (PL_AUDIO_CMD + 0x2c)

#define PL_AUDIO_CMD_I2S_DOWN       (PL_AUDIO_CMD + 0x30)
#define PL_AUDIO_CMD_I2S_UP         (PL_AUDIO_CMD + 0x34)

#define PL_AUDIO_CMD_DOWNPACK       (PL_AUDIO_CMD + 0x38)
#define PL_AUDIO_CMD_UPPACK         (PL_AUDIO_CMD + 0x3c)

#define PL_AUDIO_STATUS_OUTCOUNT    (PL_AUDIO_STATUS + 0x0e)
#define PL_AUDIO_STATUS_INCOUNT     (PL_AUDIO_STATUS + 0x0c)

/* CMD's Command Register */
#define COMMAND_3_ENBO              (1 << 7)    /* Enable audio output */
#define COMMAND_3_AMUO_MASK         (7)
#define COMMAND_2_ENBI              (1 << 7)    /* Enable audio input */
#define COMMAND_2_AMUI_MASK         (7)
#define COMMAND_1_B_RSTN            (1 << 7)    /* Audio Controller Interal Reset */
#define COMMAND_1_INTE_DNDA         (1 << 3)
#define COMMAND_1_INTE_DNSA         (1 << 2)
#define COMMAND_1_INTE_UPDA         (1 << 1)
#define COMMAND_1_INTE_UPSA         (1 << 0)
#define COMMAND_0_PAUSE             (1 << 7)
#define COMMAND_0_STOP              (1 << 6)


#define COMMAND_0_MODE_MASK         0x03        /* Mode Mask */
#define COMMAND_0_SPDIF             0x01        /* Mode selete: SPDIF */
#define COMMAND_0_AC97              0x02        /*              AC97 */
#define COMMAND_0_I2S               0x03        /*              I2S */

/* CMD's Function Register */
#define FUNCTION_3_SP_ST            (1 << 4)    /* SPDIF record starting preamble */
#define FUNCTION_3_SP_V_CHK         (1 << 3)    /* SPDIF valid bit check enable */
#define FUNCTION_3_SP_P_CHK         (1 << 2)    /* SPDIF parity bit check enable */
#define FUNCTION_3_SP_I_FS_MASK     0x03        /* SPDIF up stream frequency MASK */
#define FUNCTION_3_SP_I_FS_48K      0x00        /* frequency : 48kHz */
#define FUNCTION_3_SP_I_FS_44K      0x01        /* frequency : 44.1kHz */
#define FUNCTION_3_SP_I_FS_32K      0x02        /* frequency : 32kHz */

#define FUNCTION_1_A_RSTN           (1 << 7)    /* AC97 Codec Reset */
#define FUNCTION_1_VRA              (1 << 0)    /* AC97 Codec various frequency enable */

#define FUNCTION_0_D_SIGN           (1 << 4)    /* Signed bit invert for down stream */
#define FUNCTION_0_U_SIGN           (1 << 0)    /* Signed bit invert for up stream */


/* CMD's AC97 Codec Configuration Register */
#define I2SCONFIG_3_AUSM_MASK       (0x3 << 6)  /* up stream stereo/mono */
#define I2SCONFIG_3_AUSM_STEREO     (0x0 << 6)  /* 0x stereo */
#define I2SCONFIG_3_AUSM_MONO_LEFT  (0x2 << 6)  /* 10 mono left */
#define I2SCONFIG_3_AUSM_MONO_RIGHT (0x3 << 6)  /* 11 mono right */

#define I2SCONFIG_3_AUPS_MASK       (0x3 << 4)  /* up stream channel precision */
#define I2SCONFIG_3_AUPS_20         (0x0 << 4)  /* 00: 20bits/channel */
#define I2SCONFIG_3_AUPS_18         (0x1 << 4)  /* 01: 18bits/channel */
#define I2SCONFIG_3_AUPS_16         (0x2 << 4)  /* 10: 16bits/channel */
#define I2SCONFIG_3_AUPS_8          (0x3 << 4)  /* 11: 8bits/channel */

#define I2SCONFIG_2_ADSM_MASK       (0x3 << 6)  /* down stream stereo/mono */
#define I2SCONFIG_2_ADSM_STEREO     (0x0 << 6)  /* 00 normal (stereo) */
#define I2SCONFIG_2_ADSM_MONO       (0x1 << 6)  /* 01 mono right (slot4) */
#define I2SCONFIG_2_ADSM_MONO_RIGHT (0x1 << 6)  /* 01 mono right (slot4) */
#define I2SCONFIG_2_ADSM_MONO_LEFT  (0x2 << 6)  /* 10 mono left (slot3) */


#define I2SCONFIG_2_ADPS_MASK       (0x3 << 4)  /* down stream channel precision */
#define I2SCONFIG_2_ADPS_20         (0x0 << 4)  /* 00: 20bits/channel */
#define I2SCONFIG_2_ADPS_18         (0x1 << 4)  /* 01: 18bits/channel */
#define I2SCONFIG_2_ADPS_16         (0x2 << 4)  /* 10: 16bits/channel */
#define I2SCONFIG_2_ADPS_8          (0x3 << 4)  /* 11: 8bits/channel */


/* I2S Down Stream and Up Stream Channel */
#define I2SDOWN_DLENG_OFF           8
#define I2SDOWN_DOFFS_OFF           20
#define I2SUP_ULENG_OFF             8
#define I2SUP_UOFFS_OFF             20


#define I2S_WSO_MODE_OUTPUT                (0x00 << 4)
#define I2S_WSI_MODE_OUTPUT                (0x00 << 0)
#define I2S_WSI_MODE_INPUT         		(1<<1)    /*wsi_mode to input*/
#define I2S_WSO_MODE_INPUT         		(1<<5)   /*wso_mode to input*/

/* CMD's AC97Clock register */
#define AC97CLOCK_CLKE              (1 << 0)    /* 1: external CODEC clock, 0: internal D_CLK */
#define AC97CLOCK_CLKD              (1 << 1)    /* 1: divider enable, 0: divider disable */


/* STATUS's Register */
#define STATUS_SEMA_ENBO            (1UL << 31) /* Down stream semaphore active */
#define STATUS_ENBO                 (1UL << 27) /* Down stream active */
#define STATUS_DS_INT               (1UL << 23) /* Down stream interrupt source, read clear */
#define STATUS_DS_OB_EMP            (1UL << 19) /* Down stream buffer empty flag */
#define STATUS_DS_OB_FUL            (1UL << 18) /* Down stream buffer full flag */
#define STATUS_DS_OB_OVF            (1UL << 17) /* Down stream buffer overflow flag */
#define STATUS_DS_OB_UNF            (1UL << 16) /* Down stream buffer underflow flag */

#define STATUS_SEMA_ENBI            (1UL << 15) /* Down stream semaphore active */
#define STATUS_ENBI                 (1UL << 11) /* Down stream active */
#define STATUS_US_INT               (1UL << 7)  /* Down stream interrupt source, read clear */
#define STATUS_US_IB_EMP            (1UL << 3)  /* Down stream buffer empty flag */
#define STATUS_US_IB_FUL            (1UL << 2)  /* Down stream buffer full flag */
#define STATUS_US_IB_OVF            (1UL << 1)  /* Down stream buffer overflow flag */
#define STATUS_US_IB_UNF            (1UL << 0)  /* Down stream buffer underflow flag */

#define CONFIG_2_ADPS_16        (0x2 << 4)  /* 10: 16bits/channel */
#define CONFIG_2_ADPS_8         (0x3 << 4)  /* 11: 8bits/channel */

#define CONFIG_3_AUPS_16        (0x2 << 4)  /* 10: 16bits/channel */
#define CONFIG_3_AUPS_8         (0x3 << 4)  /* 11: 8bits/channel */

/* GPIO Pin for AUDIO */
#define C_IO_0            	(1 << 0)   /* I */
#define C_IO_1             	(1 << 1)   /* I */
#define C_IO_2            	(1 << 2)   /* O */
#define C_IO_3                	(1 << 3)   /* O */
#define C_IO_4             	 (1 << 4)   /* I */
#define GPIO_ALL_PIN        0x1f

#define DIR_OUT     1
#define DIR_IN      0

#define MCLK                12000000

#define advance_idx(idx, up)    ({(idx)++; if ((idx) >= (up << 1)) (idx) = 0; })
#define regular_idx(idx, up)    (((idx) >= (up)) ? (idx)-(up) : (idx))
#define ABS(x)                  (((x) < 0) ? -(x) : (x))


#define MAX_SLOTS                   64

#define HW_FLAG_EMP     (1 << 3)
#define HW_FLAG_FUL     (1 << 2)
#define HW_FLAG_OVF     (1 << 1)
#define HW_FLAG_UNF     (1 << 0)

#define ST_CLOSE    0
#define ST_OPEN     1
#define ST_RW       2

#define HW_DMA_COPY_BUG
#ifdef HW_DMA_COPY_BUG
#define is_in_empty(xin)    ((xin)->ihw  == (xin)->ihead || \
                             (xin)->ihw == ((((xin)->ihead+1) == ((xin)->slots<<1)) ? 0 : (xin)->ihead+1))
#else
#define is_in_empty(xin)    ((xin)->ihead == (xin)->ihw)
#endif

#define get_user_ret(x, ptr, ret) ({if (get_user(x,ptr)) return ret; })
#define validate_state(x, xstate, ret)   ({ if (x.state != xstate) return ret; })

static int aic23_write_value(int reg, u16 value);

static unsigned short	normal_i2c[]		= {TLV320AIC23_I2C_IOADDR, AUDIO_I2C_CLIENT_END};
static unsigned short	normal_i2c_range[]	= {AUDIO_I2C_CLIENT_END};
static unsigned int		normal_isa[]		= {AUDIO_ISA_END};
static unsigned int		normal_isa_range[]	= {AUDIO_ISA_END};

SENSORS_INSMOD_1(TLV320AIC23B);
static struct i2c_driver aic23_driver;
static struct i2c_client *aic23_client = NULL;

/******************************************************************************
 *  Internal use structure
 ******************************************************************************/
static int DEFAULT_OUT_BUF_SLOT_SIZE_ORDER = PAGE_SHIFT;    /* 4096 */
static int DEFAULT_OUT_BUF_SLOTS = 8;
static int DEFAULT_IN_BUF_SLOT_SIZE_ORDER = PAGE_SHIFT;     /* 4096 */
static int DEFAULT_IN_BUF_SLOTS = 8;


struct pl_i2s_info 
{
	u32     slots;
	u32     runs;       /* for output it's underrun count, for input it's overrun count */
	u32     jam;        /* 2 issues is done under a interrupt */
	u32     flags;
	u32     losts;      /* interrupt disappear */
	u32     restores;   /* count restore from hardware error */
	u32     tbytes;         /* total bytes after beginning of open */
	u32     bbytes;         /* buffered bytes */
	u32     bytes_per_sec;  /* bytes per second */
	u32     bytes_per_hour; /* bytes per hour */
	u32     ptradjust;      /* ptr adjust */
};

struct dma_mde 
{
	u32     addr;
	u32     dcnt: 21;
	u32     nxt:  8;
	u32     j:    1;
	u32     v:    1;
	u32     e:    1;
};

struct pl_i2s_buf
{
	int         state;                  /* ST_CLOSE, ST_OPEN, ST_RW */
	int         suspend;                /* 1 Device suspend now */
	int         start;
	int         slots;
	int         order;                  /* the buffer size order, the buffer size = 1 << order */
	int         ihw;                    /* index to hardware current function slots */
	int         ihead;                  /* index to next ready mde slot */
	int         itail;                  /* index to next available (free) mde slot */
	atomic_t    issues;
	int         copys;                  /* the current slots copy from/to user buffer */
	int         threshold;              /* a slot should be filled at least size to send to device */
	struct dma_mde  mde[MAX_SLOTS] __attribute__((aligned (8)));    /* #002 */
	int         hw_error;               /* once hardware underflow, driver should wait DMA stopping */
	wait_queue_head_t   wait;           /* wait for buffer ready (read) or available (write) */
	wait_queue_head_t   drain;          /* wait for buffer drain out */
	int  (*issue)(struct pl_i2s_buf *abuf);
	struct  pl_i2s_info info;
	int         alarmpid;               /* pid for alarm */
	int         alarm;                  /* alarm active flag */
};


typedef struct pl_i2s_tag 
{
	int                 dspo;           /* dsp output device */
	int                 oused;
	int                 osrate;         /* output sample rate */
	struct pl_i2s_buf   out;
	int                 dspi;           /* dsp intput device */
	int                 iused;
	int                 israte;         /* input sample rate */
	struct pl_i2s_buf   in;
	/* mixer device */
	int                 dev_mixer;
	int                 micboost;
	int                 recio;          /* 0 line in, 1 mic */
	int                 modcnt;
	int                 power;
	unsigned int        mixer_state[3];
} pl_i2s_t;

typedef struct aic23_info
{
	u16 volume_reg_left;
	u16 volume_reg_right;
	u16 input_gain_reg_left;
	u16 input_gain_reg_right;
	u16 power;			/* For POWER_DOWN_CONTROL_ADDR */
	u16 mask;			/* For ANALOG_AUDIO_CONTROL_ADDR */
	int mic_loopback;
	int mic_enable;
	int sta;
	int power_down;
	int initialized;
} aic23_info_t;

static pl_i2s_t pl_i2s_ins;
static aic23_info_t aic23_info_l;
static int selftest;

#define AUDIO_VOLUME_TABLEN 21
static const unsigned short HPVOLUME_TAB[AUDIO_VOLUME_TABLEN] = 
{
	/* LEVEL: 0 ,  Mute -73dB */
	 0x30,
	 
	/* LEVEL:  1 ~  9 */
	 0x37, 0x3D, 0x43, 0x49, 0x4F,0x55, 0x5B, 0x61, 0x67,
	 
	/* LEVEL: 10 ~ 13 */
	 0x71, 0x73, 0x75, 0x77,
	 
	/* LEVEL: 14 , Normal: 0 dB */
	 0x79,			
	 
	/* LEVEL: 15 ~ 20 */
	 0x7A, 0x7B, 0x7C, 0x7D, 0x7E,

	/* LEVEL: 21, MAX: 6 dB*/
	 0x7F
};

/******************************************************************************
 *  downstream and upstream buffer mantain functions
 ******************************************************************************/
/**
 *      ihw ihead              itail
 *       v   v                   v
 *  +-------------------------------------+
 *  | | |*|*|.|.|.|.|.|.|.|.|.|.| | | | | |
 *  +-------------------------------------+
 *    issueing  ready but not issue
 */
static spinlock_t  issue_lock = SPIN_LOCK_UNLOCKED;
static int issue_out(struct pl_i2s_buf *out)
{
	int flags = 0;
	int idx;
	// int l;
	struct pl_i2s_info *out_info = &out->info;

	if (get_dma_residue(DMA_PL_AC97_OUT) == 0) 
	{
		if (out->hw_error) 
		{
			out->hw_error = 0;
			out_info->restores++;
		}
		// out_info->runs++;
		wake_up(&out->drain);
		AUDIO_LINE;
	}

	if (!out->start)               /* forbidden by PAUSE or TRIGGER */
		return 4;

	if (out->ihead == out->itail)  /* EMPTY, no more ready slot */
	/* inactive dma mechine */
		return 1;

	/* check hardware error flag */
	if (out->hw_error)
		return 3;

	/* check hardware busy, if hw busy then just leaveing */
	if (atomic_read(&out->issues) >= 2)
		return 2;

	/**** Critical Section ****/
	if (! in_interrupt())
		spin_lock_irqsave(&issue_lock, flags);

	/* it should be checked agian before issuing */
	if (out->ihead != out->itail &&
	!out->hw_error && atomic_read(&out->issues) < 2) 
	{
		idx = regular_idx(out->ihead, out->slots);

		if (out->mde[idx].dcnt < 16)
			out->mde[idx].dcnt = 16;


		pci_map_single(NULL, bus_to_virt(out->mde[idx].addr), out->mde[idx].dcnt, PCI_DMA_TODEVICE);
		pci_map_single(NULL, out->mde, MAX_SLOTS * sizeof(struct dma_mde), PCI_DMA_TODEVICE);
		writew(out->mde[idx].dcnt >> 2, PL_AUDIO_CMD_DOWNPACK);
		
#if	1
{
		static unsigned int sdcnt = 0;
		int                 dcount = 0;
		unsigned int        enbo_start;

		while (dcount < 1000) 
		{
			enbo_start = readb(PL_AUDIO_CMD_COMMAND_3);
			sdcnt = readw(PL_AUDIO_STATUS_OUTCOUNT);
			if ((enbo_start & 0x80) && sdcnt <= 2)
			{
				AUDIO_DBG("...hw error\n");
				dcount++;
				udelay(25);
			} 
			else 
			{
				break;
			}
		}
}
#endif
		out->issues.counter++;
		writeb(COMMAND_3_ENBO | 2, PL_AUDIO_CMD_COMMAND_3);

		/* release a mde slot */
		advance_idx(out->ihead, out->slots);

		/* active dma machine */
		if (get_dma_residue(DMA_PL_AC97_OUT) == 0) 
		{
			AUDIO_LINE;
			set_dma_mde_idx(DMA_PL_AC97_OUT, idx);
			enable_dma(DMA_PL_AC97_OUT);
			out_info->runs++;
		}
	}

	if (! in_interrupt())
	spin_unlock_irqrestore(&issue_lock, flags);
	/**** End Critical Section ****/


	return 0;
}


/**
 *      ihead               ihw itail
 *       v                   v   v
 *  +-------------------------------------+
 *  | | |.|.|.|.|.|.|.|.|.|.|*|*| | | | | |
 *  +-------------------------------------+
 *       ready for copying     issuing
 */


static int issue_in(struct pl_i2s_buf *in)
{
	int flags = 0;
	int idx;
	int i;
	struct pl_i2s_info *in_info = &in->info;

	if (get_dma_residue(DMA_PL_AC97_IN) == 0 || atomic_read(&in->issues) == 0) 
	{
		if (in->hw_error) 
		{
			in->hw_error = 0;
			in_info->restores++;
		}
		// in_info->runs++;
		wake_up(&in->drain);
	}

	if (!in->start) /* forbidden by PAUSE or TRIGGER */
		return 4;

	if (ABS(in->itail-in->ihead) == in->slots)  /* FULL, no more available slot */
		return 1;

	/* check hardware error flag */
	if (in->hw_error)
		return 3;

	if (atomic_read(&in->issues) >= 2)
		return 2;

	/**** Critical Section ****/
	if (! in_interrupt())
		spin_lock_irqsave(&issue_lock, flags);

	for (i = 0; i < 2 -atomic_read(&in->issues); i++)
	{
		/* it should be check again before issuing */
		if (ABS(in->itail-in->ihead) != in->slots && !in->hw_error &&
		atomic_read(&in->issues) < 2) 
		{
			idx = regular_idx(in->itail,in->slots);
			// pci_map_single(NULL, bus_to_virt(in->mde[idx].addr), in->mde[idx].dcnt, PCI_DMA_TODEVICE);
			pci_map_single(NULL, in->mde, MAX_SLOTS * sizeof(struct dma_mde), PCI_DMA_TODEVICE);

			writew(in->mde[idx].dcnt >> 2, PL_AUDIO_CMD_UPPACK);
			in->issues.counter++;
			writeb(COMMAND_2_ENBI | 2, PL_AUDIO_CMD_COMMAND_2);
			/* release a mde slot */
			advance_idx(in->itail, in->slots);

			/* active dma machine */
			if (get_dma_residue(DMA_PL_AC97_IN) == 0)
			{
				set_dma_mde_idx(DMA_PL_AC97_IN, idx);
				enable_dma(DMA_PL_AC97_IN);
				in_info->runs++;
			}
		}
	}

	if (! in_interrupt())
	spin_unlock_irqrestore(&issue_lock, flags);
	/**** End Critical Section ****/


	return 0;
}


static int drain_out(struct pl_i2s_buf *out)
{
	int idx;

	/* flush the last slot */
	idx = regular_idx(out->itail, out->slots);
	if (out->copys > 0) 
	{
		out->mde[idx].dcnt = out->copys & ~3;  /* aligned to dword size */
	} 
	else 
	{
		out->mde[idx].dcnt = 16;    /* add a junk line */
		memset((char *)bus_to_virt(out->mde[idx].addr), 0, 16);
	}
	out->copys = 0;
	out->mde[idx].v = 1;    /* valid mde entry */
	out->mde[idx].e = 1;    /* mark last mde */
	advance_idx(out->itail, out->slots);
	issue_out(out);

	//wait_event(out->drain, get_dma_residue(DMA_PL_AC97_OUT) == 0);
	wait_event_interruptible(out->drain, get_dma_residue(DMA_PL_AC97_OUT) == 0);

	writeb(2, PL_AUDIO_CMD_COMMAND_3);  /* clear DMAD flag */
	return 0;
}

static int drain_in(struct pl_i2s_buf *in)
{
	writeb(0, PL_AUDIO_CMD_COMMAND_2);
	wait_event(in->drain, get_dma_residue(DMA_PL_AC97_IN) == 0);


	writeb(0, PL_AUDIO_CMD_COMMAND_2);
	return 0;
}

static void inline start_issue(struct pl_i2s_buf *abuf)
{
	abuf->start = 1;
	if (abuf->state == ST_RW)
		abuf->issue(abuf);
}

static void inline stop_issue(struct pl_i2s_buf *abuf)
{
	abuf->start = 0;
}

static void pause_issue(struct pl_i2s_buf *abuf)
{
	if (abuf->start)
	    stop_issue(abuf);
	else
	    start_issue(abuf);
}

static void inline clean_buffer(struct pl_i2s_buf *abuf)
{
	abuf->ihw = abuf->itail = abuf->ihead = 0;
	atomic_set(&abuf->issues, 0);
	abuf->copys = 0;
	abuf->hw_error = 0;
}

/******************************************************************************
 *  interrupt handler
 ******************************************************************************/
static void pl_i2s_handler(int irq, void *dev_id, struct pt_regs* regs)
{
	pl_i2s_t *pl_i2s = dev_id;
	struct pl_i2s_buf *out = &pl_i2s->out;
	struct pl_i2s_buf *in = &pl_i2s->in;
	struct pl_i2s_info *out_info = &out->info;
	struct pl_i2s_info *in_info = &in->info;
	int hw_out = 0, hw_in = 0;
	int i;
	u32 status = readl(PL_AUDIO_STATUS);
#ifdef DEBUG
	int oissues = atomic_read(&out->issues);
	int iissues = atomic_read(&in->issues);
#endif

	if (atomic_read(&out->issues) == 2)
	{
		if ((status & STATUS_SEMA_ENBO) == 0) 
		{
			/* codec has comsumed a down stream slot */
			out->issues.counter--;
			hw_out++;
			out->mde[regular_idx(out->ihw,out->slots)].v = 0;   /* invalid mde entry */
			out_info->bbytes -= out->mde[regular_idx(out->ihw, out->slots)].dcnt;
			advance_idx(out->ihw, out->slots);
		}
	}

	if (atomic_read(&out->issues) == 1) 
	{
		if ((status & STATUS_ENBO) == 0)
		{
			/* codec has comsumed a down stream slot */
			out->issues.counter--;
			hw_out++;
			out->mde[regular_idx(out->ihw,out->slots)].v = 0;   /* invalid mde entry */
			out_info->bbytes -= out->mde[regular_idx(out->ihw, out->slots)].dcnt;
			advance_idx(out->ihw, out->slots);
		}
	}

	if (atomic_read(&in->issues) == 2) 
	{
		if ((status & STATUS_SEMA_ENBI) == 0) 
		{
			/* codec has comsumed a up stream slot */
			in->issues.counter--;
			hw_in++;
			in_info->bbytes += in->mde[regular_idx(in->ihw, in->slots)].dcnt;
			advance_idx(in->ihw, in->slots);
		}
	}

	if (atomic_read(&in->issues) == 1)
	{
		if ((status & STATUS_ENBI) == 0) 
		{
			/* codec has comsumed a up stream slot */
			in->issues.counter--;
			hw_in++;
			in_info->bbytes += in->mde[regular_idx(in->ihw, in->slots)].dcnt;
			advance_idx(in->ihw, in->slots);
		}
	}

#ifdef DEBUG

	printk("oissues %d (%d) iissues %d (%d) status = %08x hw_out %d (%d %d %d) hw_in %d (%d %d %d)\n",
	oissues, atomic_read(&out->issues),
	iissues, atomic_read(&in->issues),
	status, hw_out,
	out->ihw, out->ihead, out->itail,
	hw_in, in->ihead, in->ihw, in->itail);
#endif

	if (status & (STATUS_DS_OB_OVF | STATUS_DS_OB_UNF))
		out->hw_error++;

	if (status & (STATUS_US_IB_OVF | STATUS_US_IB_UNF))
		in->hw_error++;

	for (i = 0; i < hw_out; i++)
	{
		issue_out(out);
		wake_up_interruptible(&out->wait);
	}

	/* kill sigalarm */
	if (out->alarm && hw_out)
		kill_proc(out->alarmpid, SIGALRM, 1);


	for (i = 0; i < hw_in; i++)
	{
		issue_in(in);
		wake_up_interruptible(&in->wait);
	}

	/* calculate statistic information */
	if (hw_out == 2)
		out_info->jam++;

	if (hw_out && (status & STATUS_DS_INT) == 0)
		out_info->losts++;

	out_info->slots += hw_out;
	out_info->flags |= (status >> 16) & 0x0f;

	if (hw_in == 2)
		in_info->jam++;

	if (hw_in && (status & STATUS_US_INT) == 0)
		in_info->losts++;


	in_info->slots += hw_in;
	in_info->flags |= status & 0x0f;

}

/******************************************************************************
 *  Configure functions
 ******************************************************************************/
static void pl_dsp_set_channels(int channels, int output)
{
	u8 data;

	if (output) 
	{
		data = readb(PL_AUDIO_CMD_CONFIG_2) & ~I2SCONFIG_2_ADSM_MASK;
		
		if (channels == 2)
			writeb(data | I2SCONFIG_2_ADSM_STEREO, PL_AUDIO_CMD_CONFIG_2);
		else
			writeb(data | I2SCONFIG_2_ADSM_MONO, PL_AUDIO_CMD_CONFIG_2);
	} 
	else 
	{
		data = readb(PL_AUDIO_CMD_CONFIG_3) & ~I2SCONFIG_3_AUSM_MASK;
		
		if (channels == 2)
			writeb(data | I2SCONFIG_3_AUSM_STEREO, PL_AUDIO_CMD_CONFIG_3);
		else
			writeb(data | I2SCONFIG_3_AUSM_MONO_LEFT, PL_AUDIO_CMD_CONFIG_3);
	}
}


static int pl_dsp_get_channels(int output)
{
	int channels;
	u8 data;

	if (output) 
	{
		data = readb(PL_AUDIO_CMD_CONFIG_2) & I2SCONFIG_2_ADSM_MASK;
		if (data == I2SCONFIG_2_ADSM_STEREO)
			channels = 2;
		else
			channels = 1;
	} 
	else 
	{
		data = readb(PL_AUDIO_CMD_CONFIG_3) & I2SCONFIG_3_AUSM_MASK;
		if (data == I2SCONFIG_3_AUSM_STEREO)
			channels = 2;
		else
			channels = 1;
	}

	return channels;
}


static inline int is_valid_rate(u16 rate)
{
	if (rate == 48000 || rate == 44100 || rate == 32000 || rate == 8000)
		return 1;

	return 0;
}


static u16 pl_dsp_set_samplerate(pl_i2s_t *pl_i2s, u16 rate, int output)
{
	int  leng;
	u16 val;

	AUDIO_DBG("rate = %d\n", rate);
	if (!is_valid_rate(rate))
	{
		rate = pl_i2s->osrate;
	}
	else
	{
		switch (rate) 
		{
			case 8000:
				val = SR1SR | SR0SR|USB_CLK_ON;
				break;
				
			case 16000:
				val = CLKIN_HALF| SR2SR | SR1SR|USB_CLK_ON;
				break;
				
			case 22050:
				val = CLKIN_HALF| SR3SR | BOSR_384fs|USB_CLK_ON;
				break;
				
			case 24000:
				val = CLKIN_HALF|USB_CLK_ON;
				break;
				
			case 32000:
				val =  SR2SR | SR1SR|USB_CLK_ON ;
				break;
				
			case 44100:
				val =  SR3SR | BOSR_384fs|USB_CLK_ON;
				break;
				
			case 48000:
				val = USB_CLK_ON;
				break;
				
			default:
				val = USB_CLK_ON;
				break;
		}
		AUDIO_DBG("rate val = %d\n",val);
		aic23_write_value(SAMPLE_RATE_CONTROL_ADDR, val);
		
		if (output) 
		{
           		 /* set I2S Down stream channel length register */
			leng = (MCLK/rate) >> 1;	 
			writel(leng << I2SDOWN_DLENG_OFF, PL_AUDIO_CMD_I2S_DOWN);
			pl_i2s->osrate = rate;
			AUDIO_LINE;
	
		} 
		else 
		{
			/* set I2S Up stream channel length register */
			leng = (MCLK/rate) >> 1;
			writel(leng << I2SUP_ULENG_OFF, PL_AUDIO_CMD_I2S_UP);
			pl_i2s->israte = rate;
			AUDIO_LINE;
		}
		
	}	
			
	return rate;
}

static u32 pl_dsp_get_samplerate(pl_i2s_t *pl_i2s, int output)
{
	u16 rate;
	if (output)
	{
		rate = pl_i2s->osrate;
	}
	else 
	{
		rate = pl_i2s->israte;
	}

	return rate;
}

static int pl_dsp_get_format(int output)
{
	int unsign = 0;
	int fmt;
	u8 data;

	if (output)
	{
		unsign = readb(PL_AUDIO_CMD_FUNCTION_0) & FUNCTION_0_D_SIGN;
		data = readb(PL_AUDIO_CMD_CONFIG_2) & I2SCONFIG_2_ADPS_MASK;
		fmt = (data == I2SCONFIG_2_ADPS_16) ? 16 : 8;
	} 
	else 
	{
		unsign = readb(PL_AUDIO_CMD_FUNCTION_0) & FUNCTION_0_U_SIGN;
		data = readb(PL_AUDIO_CMD_CONFIG_3) & I2SCONFIG_3_AUPS_MASK;
		fmt = (data == I2SCONFIG_3_AUPS_16) ? 16 : 8;
	}

	if (!unsign) 
	{
		if (fmt == 8)
			return AFMT_S8;
		else
			return AFMT_S16_LE;
	} 
	else 
	{
		if (fmt == 8)
			return AFMT_U8;
		else
			return AFMT_U16_LE;
	}

}

static int pl_dsp_set_format(int format, int output)
{
	int sign;
	int fmt;
	u8 data;

	switch(format)
	{
		case AFMT_QUERY:
			return pl_dsp_get_format(output);

		case AFMT_U8:
			sign = 0; fmt = 8; 
			break;

		case AFMT_S8:
			sign = 1; fmt = 8; 
			break;

		case AFMT_U16_LE:
			sign = 0; fmt = 16; 
			break;

		case AFMT_S16_LE:
		default:
			sign = 1; fmt = 16;
			format = AFMT_S16_LE;
			break;
	}

	if (output)
	{
		data = readb(PL_AUDIO_CMD_FUNCTION_0) & ~FUNCTION_0_D_SIGN;

		if (sign)
			writeb(data, PL_AUDIO_CMD_FUNCTION_0);
		else
			writeb(data | FUNCTION_0_D_SIGN, PL_AUDIO_CMD_FUNCTION_0);

		data = readb(PL_AUDIO_CMD_CONFIG_2) & ~I2SCONFIG_2_ADPS_MASK;

		if (fmt == 8)
			writeb(data | I2SCONFIG_2_ADPS_8, PL_AUDIO_CMD_CONFIG_2);
		else
			writeb(data | I2SCONFIG_2_ADPS_16, PL_AUDIO_CMD_CONFIG_2);
	} 
	else 
	{
		data = readb(PL_AUDIO_CMD_FUNCTION_0) & ~FUNCTION_0_U_SIGN;

		if (sign)
			writeb(data, PL_AUDIO_CMD_FUNCTION_0);
		else
			writeb(data | FUNCTION_0_D_SIGN, PL_AUDIO_CMD_FUNCTION_0);

		data = readb(PL_AUDIO_CMD_CONFIG_3) & ~I2SCONFIG_3_AUPS_MASK;

		if (fmt == 8)
			writeb(data | I2SCONFIG_3_AUPS_8, PL_AUDIO_CMD_CONFIG_3);
		else
			writeb(data | I2SCONFIG_3_AUPS_16, PL_AUDIO_CMD_CONFIG_3);
	}

	return format;
}

static void pl_calc_time_base(pl_i2s_t *pl_i2s, int output)
{
	int ch;
	int samples;
	int fmt;
	int bytes_per_sample;
	struct pl_i2s_info *info;

	ch = pl_dsp_get_channels(output);
	samples = pl_dsp_get_samplerate(pl_i2s, output);
	fmt = pl_dsp_get_format(output);
	
	if (fmt == AFMT_S16_LE || fmt == AFMT_U16_LE)
		bytes_per_sample = 2;
	else
		bytes_per_sample = 1;
	
	if (output)
		info = &pl_i2s->out.info;
	else
		info = &pl_i2s->in.info;

	info->bytes_per_sec = ch*bytes_per_sample*samples;
	info->bytes_per_hour = info->bytes_per_sec * 60 * 60;

}

static int pl_dsp_getodelay(struct pl_i2s_buf *out)
{
	int size;
	size = readw(PL_AUDIO_STATUS_OUTCOUNT) << 2;
	size = out->info.bbytes - size;
	if (size < 0)
		size = 0;

	return size;
}

static int pl_dsp_getidelay(struct pl_i2s_buf *in)
{
	int size;

	size = readw(PL_AUDIO_STATUS_INCOUNT) << 2;
	size = in->info.bbytes + size;
	if (size < 0)
		size = 0;

	return size;
}

static void pl_dsp_getoptr(struct pl_i2s_buf *out, count_info *info)
{
	info->bytes = out->info.tbytes;
	info->blocks = out->info.slots;
	out->info.slots = 0;
	info->ptr = out->info.bbytes;
}

static void pl_dsp_getiptr(struct pl_i2s_buf *in, count_info *info)
{
	info->bytes = in->info.tbytes;
	info->blocks = in->info.slots;
	in->info.slots = 0;
	info->ptr = in->info.bbytes;
}

static int pl_dsp_setfragment(struct pl_i2s_buf *abuf, int valid, int nfrag, int frag_ord)
{
	int size;
	int i;
	u32 p = 0;

	if (nfrag > 64 || nfrag < 4)
		return -EINVAL;

	if (frag_ord > 15 || frag_ord < 8)
		return -EINVAL;

	/* don't change anything, if order, fragmets is same as original */
	if (abuf->order == frag_ord && abuf->slots == nfrag)
		return 0;


	/* release allocated buffer */
	if (abuf->order != 0) 
	{
		if (abuf->order <= PAGE_SHIFT) 
		{
			for (i = 0; i < abuf->slots; i+= (1 << (PAGE_SHIFT-abuf->order))) 
			{
				free_page((u32)bus_to_virt(abuf->mde[i].addr));
			}
		} 
		else 
		{
			for (i = 0; i < abuf->slots; i++) 
			{
			free_pages((u32)bus_to_virt(abuf->mde[i].addr), abuf->order - PAGE_SHIFT);
			}
		}
		abuf->order = 0;
	}

	/* allocate new buffer */
	size = 1 << frag_ord;
	for (i = 0; i < nfrag; i++) 
	{
		if (frag_ord <= PAGE_SHIFT) 
		{
			if ((i & ((1 << (PAGE_SHIFT-frag_ord)) - 1)) == 0) 
			{
				p = abuf->mde[i].addr = (u32)virt_to_bus((void *)__get_free_page(GFP_KERNEL));
			} 
			else 
			{
				p += (1 << frag_ord);
				abuf->mde[i].addr = p;
			}
		}
		else 
		{
			abuf->mde[i].addr = (u32) virt_to_bus((void *)__get_free_pages(GFP_KERNEL, frag_ord -PAGE_SHIFT));
		}
		abuf->mde[i].nxt = i+1;
		abuf->mde[i].e = 0;
		abuf->mde[i].v = valid;
		abuf->mde[i].dcnt = size;
	}

	abuf->mde[nfrag-1].nxt = 0;
	abuf->slots = nfrag;
	abuf->order = frag_ord;
	abuf->threshold = 1 << frag_ord;

	for (i = 0; i < abuf->slots; i++)
	{
		if (abuf->mde[i].addr == 0) 
		{
			abuf->order = 0;
			return -ENOMEM;
		}
	}

	return 0;
}

static int get_order_by_size(int size)
{
	int i;

	for (i = 0; i < 30; i++)
	{
		if ((1 << i) >= size)
			break;
	}

	return i;
}


/******************************************************************************
 *  DSP functions
 ******************************************************************************/

static int pl_dsp_open(struct inode *inode, struct file *file)
{
	int fmt;
	int minor = MINOR(inode->i_rdev);

	if ((minor & 0xf) == SND_DEV_DSP) 
	{
		fmt = AFMT_U8;
	} else if ((minor & 0xf) == SND_DEV_DSP16)
	{
		fmt = AFMT_S16_LE;
	} else
		return -ENODEV;


	if ((file->f_mode & FMODE_WRITE) && pl_i2s_ins.out.state != ST_CLOSE)
		return -EBUSY;

	if ((file->f_mode & FMODE_READ) && pl_i2s_ins.in.state != ST_CLOSE)
		return -EBUSY;


	/* downstream */
	if (file->f_mode & FMODE_WRITE) 
	{
		pl_i2s_ins.out.state = ST_OPEN;
		pl_i2s_ins.out.start = 1;
		aic23_power_up();
		pl_dsp_set_channels(2, DIR_OUT);
		pl_dsp_set_format(fmt, DIR_OUT);
		pl_dsp_set_samplerate(&pl_i2s_ins,8000, DIR_OUT);
		memset(&pl_i2s_ins.out.info, 0, sizeof(struct pl_i2s_info));
		pl_calc_time_base(&pl_i2s_ins, DIR_OUT);
	}


	/* upstream */
	if (file->f_mode & FMODE_READ) 
	{
		pl_i2s_ins.in.state = ST_OPEN;
		pl_i2s_ins.in.start = 1;
		aic23_power_up();
		pl_dsp_set_channels(2, DIR_IN);
		pl_dsp_set_format(fmt, DIR_IN);
		pl_dsp_set_samplerate(&pl_i2s_ins,8000, DIR_IN);
		memset(&pl_i2s_ins.in.info, 0, sizeof(struct pl_i2s_info));
		pl_calc_time_base(&pl_i2s_ins, DIR_IN);
	}


	file->private_data = &pl_i2s_ins;

	return 0;
}

static int pl_dsp_release(struct inode *inode, struct file *file)
{
	int rc1 = 0, rc2 = 0;
	pl_i2s_t *pl_i2s = (pl_i2s_t *)file->private_data;

	AUDIO_LINE;
	if (file->f_mode & FMODE_WRITE) 
	{
		rc1 = drain_out(&pl_i2s->out);
		pl_i2s->out.state = ST_CLOSE;
		clean_buffer(&pl_i2s->out);
	}

	if (file->f_mode & FMODE_READ)
	{
		rc2 = drain_in(&pl_i2s->in);
		pl_i2s->in.state = ST_CLOSE;
		clean_buffer(&pl_i2s->in);

	}

	if((ST_CLOSE ==  pl_i2s->in.state )&&( ST_CLOSE == pl_i2s->out.state))
	{
		aic23_power_down();
	}

	if (rc1)
		return rc1;
	else
		return rc2;
}

static ssize_t pl_dsp_write(struct file *file, const char *buffer, size_t count, loff_t *ppos)
{
	int nwrite = 0;
	pl_i2s_t *pl_i2s = (pl_i2s_t *)file->private_data;
	struct pl_i2s_buf *out;
	int len, idx;
	int rc, bufsize;
	const char *p;


	if (pl_i2s == NULL)
		return -EINVAL;

	out = &pl_i2s->out;

	if (ppos != &file->f_pos)
		return -ESPIPE;

	if (!access_ok(VERIFY_READ, buffer, count))
		return -EFAULT;

	count = count & ~3; /* trim to dword size */

	if ((count & 3) != 0)   /* should be double word size, limit by ac97 controller */
		return -EINVAL;

	out->state = ST_RW;

	start_issue(out);    

	bufsize = 1 << out->order;
	p = buffer;
	
	while (count > 0)
	{
		if (ABS(out->itail-out->ihw) == out->slots)  /* FULL, no more avaliable slot */
		{
			if (file->f_flags & O_NONBLOCK)
				return (nwrite == 0) ? -EAGAIN : nwrite;
			else
			{
				AUDIO_LINE;
				rc = wait_event_interruptible(out->wait, ABS(out->itail-out->ihw)!=out->slots);
				if (rc < 0)
					return rc;
				AUDIO_LINE;
			}
		}

#ifdef DEBUG
		printk("out->issues %d (%d %d %d)\n", atomic_read(&out->issues), out->ihw, out->ihead, out->itail);
#endif
		idx = regular_idx(out->itail, out->slots);
		len = (count > (bufsize-out->copys)) ? (bufsize-out->copys) : count;
		copy_from_user((void *)bus_to_virt(out->mde[idx].addr)+out->copys, (void *)p, len);
		out->copys += len;
		count -= len;
		out->info.bbytes += len;
		out->info.tbytes += len;
		p += len;
		nwrite += len;
		if (out->info.tbytes >= out->info.bytes_per_hour) 
		{
			AUDIO_LINE;
			out->info.ptradjust++;
			out->info.tbytes -= out->info.bytes_per_hour;
		}

		if (out->copys >= out->threshold) 
		{
			AUDIO_LINE;
			out->mde[idx].dcnt = out->copys;
			out->copys = 0;
			out->mde[idx].v = 1;    /* valid mde entry */
			out->mde[idx].e = 0;
			advance_idx(out->itail, out->slots);
			issue_out(out);
		}
	}

	return nwrite;
}

static ssize_t pl_dsp_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	int nread = 0;
	pl_i2s_t *pl_i2s = (pl_i2s_t *)file->private_data;
	struct pl_i2s_buf *in;
	int len, idx;
	int rc, bufsize;
	char *p;


	if (pl_i2s == NULL)
		return -EINVAL;


	in = &pl_i2s->in;

	if (ppos != &file->f_pos)
		return -ESPIPE;


	if (!access_ok(VERIFY_WRITE, buffer, count))
		return -EFAULT;

	in->state = ST_RW;
	bufsize = 1 << in->order;
	p = buffer;
	issue_in(in);


	while(count > 0) 
	{

		if (is_in_empty(in))  /* EMPTY, no more data for copying */
		{
			if (file->f_flags & O_NONBLOCK)
				return (nread == 0) ? -EAGAIN: nread;
			else 
			{
				rc = wait_event_interruptible(in->wait, !is_in_empty(in));
				if (rc < 0)
					return rc;
			}
		}
#ifdef DEBUG
		printk("in->issues %d (%d %d %d)\n", atomic_read(&in->issues), in->ihead, in->ihw, in->itail);
#endif

		idx = regular_idx(in->ihead, in->slots);
		len = (count > (bufsize-in->copys)) ? (bufsize-in->copys) : count;
		pci_dma_sync_single(NULL, in->mde[idx].addr, in->mde[idx].dcnt, PCI_DMA_TODEVICE);
		copy_to_user((void *)p, (void *)bus_to_virt(in->mde[idx].addr) + in->copys, len);
		count -= len;
		nread += len;
		in->copys += len;
		in->info.bbytes -= len;
		in->info.tbytes += len;
		p += len;
		if (in->info.tbytes >= in->info.bytes_per_hour) 
		{
			in->info.ptradjust++;
			in->info.tbytes -= in->info.bytes_per_hour;
		}


		if (in->copys >= in->threshold) 
		{
			advance_idx(in->ihead, in->slots);
			in->copys = 0;
			issue_in(in);
		}
	}

	return nread;
}

static unsigned int pl_dsp_poll(struct file *file, poll_table *wait)
{
	pl_i2s_t *pl_i2s = file->private_data;
	unsigned int mask = 0;

	if (file->f_mode & FMODE_WRITE)
		poll_wait(file, &pl_i2s->out.wait, wait);

	if (file->f_mode & FMODE_READ)
		poll_wait(file, &pl_i2s->in.wait, wait);


	/* is not FULL? slot available for writing */
	if ((file->f_mode & FMODE_WRITE) &&
	(ABS(pl_i2s->out.itail-pl_i2s->out.ihw) != pl_i2s->out.slots))
	{
		mask |= (POLLOUT | POLLWRNORM);
	}

	/* is not EMPTY? data is ready for copy */
	if ((file->f_mode & FMODE_READ) &&
	(!is_in_empty(&pl_i2s->in) ||
	(pl_i2s->in.ihead == pl_i2s->in.itail)))
	{
		mask |= (POLLIN | POLLRDNORM);
	}

	return mask;
}

static int pl_dsp_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int rc = 0;
	int val = 0;
	int channels = 1;
	pl_i2s_t *pl_i2s = (pl_i2s_t *)file->private_data;
	audio_buf_info info;
	count_info cinfo;

	switch(cmd) 
	{
		case OSS_GETVERSION:
		{
			rc = put_user(SOUND_VERSION, (int *)arg);
			
			break;
		}
		
		case SNDCTL_DSP_GETFMTS:
		{
			rc = put_user( (AFMT_U8 | AFMT_S8 | AFMT_S16_LE | AFMT_U16_LE), (int *)arg);
			
			break;
		}
		
		case SNDCTL_DSP_SETFMT:
		{
			get_user_ret(val, (int *)arg, -EFAULT);
			if (file->f_mode & FMODE_WRITE) 
			{
				validate_state(pl_i2s->out, ST_OPEN, -EINVAL);
				val = pl_dsp_set_format(val, DIR_OUT);
				pl_calc_time_base(pl_i2s, DIR_OUT);
			}
			if (file->f_mode & FMODE_READ) 
			{
				validate_state(pl_i2s->in, ST_OPEN, -EINVAL);
				val = pl_dsp_set_format(val, DIR_IN);
				pl_calc_time_base(pl_i2s, DIR_IN);
			}
			rc = put_user(val, (int *)arg);
			
			break;
		}
		
		case SNDCTL_DSP_CHANNELS:   /* 1 == mono, 2 == stereo */
		{
			get_user_ret(val, (int *)arg, -EFAULT);
			
			if (val != 2 && val != 1)
				return -EINVAL;
			channels = val;		
			/* go through to next case */

		}
		
		case SNDCTL_DSP_STEREO:     /* nonzero == stereo */
		{
			if (cmd == SNDCTL_DSP_STEREO) 
			{
				get_user_ret(val, (int *)arg, -EFAULT);
				channels = (val) ? 2 : 1;
			}
			
			if (file->f_mode & FMODE_WRITE) 
			{
				validate_state(pl_i2s->out, ST_OPEN, -EINVAL);
				pl_dsp_set_channels(channels, DIR_OUT);
				pl_calc_time_base(pl_i2s, DIR_OUT);
			}
			
			if (file->f_mode & FMODE_READ) 
			{
				validate_state(pl_i2s->in, ST_OPEN, -EINVAL);
				pl_dsp_set_channels(channels, DIR_IN);
				pl_calc_time_base(pl_i2s, DIR_IN);
			}

			rc = put_user(val, (int *)arg);
			
			break;
		}

		case SNDCTL_DSP_SPEED:
		{
			get_user_ret(val, (int *)arg, -EFAULT);
			if (val < 0)
				return -EINVAL;
			
			if (file->f_mode & FMODE_WRITE) 
			{
				validate_state(pl_i2s->out, ST_OPEN, -EINVAL);
				val = pl_dsp_set_samplerate(pl_i2s, val, DIR_OUT);
				pl_calc_time_base(pl_i2s, DIR_OUT);
			}
			if (file->f_mode & FMODE_READ)
			{
				validate_state(pl_i2s->in, ST_OPEN, -EINVAL);
				val = pl_dsp_set_samplerate(pl_i2s, val, DIR_IN);
				pl_calc_time_base(pl_i2s, DIR_IN);
			}

			rc = put_user(val, (int *)arg);
			
			break;
		}
		
		case SNDCTL_DSP_SYNC:
		{
			if (file->f_mode & FMODE_WRITE) 
			{
				drain_out(&pl_i2s->out);
				pl_i2s->out.state = ST_OPEN;
			}
			
			break;
		}

		case SNDCTL_DSP_SETDUPLEX:
			break;
			
		case SNDCTL_DSP_RESET:
		{
			if (file->f_mode & FMODE_READ && pl_i2s->in.state != ST_OPEN) 
			{  
				validate_state(pl_i2s->in, ST_RW, -EINVAL);
				stop_issue(&pl_i2s->in);
				writeb(0, PL_AUDIO_CMD_COMMAND_2);  /* force hardware to reset downstream */
				drain_in(&pl_i2s->in);
				clean_buffer(&pl_i2s->in);
				pl_i2s->in.state = ST_OPEN;
			}
			
			if (file->f_mode & FMODE_WRITE && pl_i2s->out.state != ST_OPEN) 
			{ 
				validate_state(pl_i2s->out, ST_RW, -EINVAL);
				stop_issue(&pl_i2s->out);
				writeb(0, PL_AUDIO_CMD_COMMAND_3);  /* force hardware to reset downstream */
				drain_out(&pl_i2s->out);
				clean_buffer(&pl_i2s->out);
				pl_i2s->out.state = ST_OPEN;
			}
			
			break;
		}
		
		case SNDCTL_DSP_POST:
		{
			if (file->f_mode & FMODE_WRITE)
				pause_issue(&pl_i2s->out);
			
			break;
		}

		case SNDCTL_DSP_GETOSPACE:
		{
			if (!(file->f_mode & FMODE_WRITE))
				return -EINVAL;

			if (pl_i2s->out.ihw == pl_i2s->out.itail)
			{
				info.fragments = pl_i2s->out.slots;
			}
			else 
			{
				info.fragments = regular_idx(pl_i2s->out.ihw, pl_i2s->out.slots) -
				regular_idx( pl_i2s->out.itail, pl_i2s->out.slots);
				if (info.fragments < 0)
					info.fragments = pl_i2s->out.slots + info.fragments;
			}

			info.fragstotal = pl_i2s->out.slots;
			info.fragsize = 1 << pl_i2s->out.order;
			info.bytes = info.fragments * info.fragsize;
			rc = copy_to_user((void *)arg, &info, sizeof(audio_buf_info));
			
			break;
		}

		case SNDCTL_DSP_GETISPACE:
		{
			if (!(file->f_mode & FMODE_READ))
				return -EINVAL;

			if (pl_i2s->in.ihw == pl_i2s->in.ihead)
			{
				info.fragments = 0;
			}
			else
			{
				info.fragments = regular_idx(pl_i2s->in.ihw, pl_i2s->in.slots) -
				regular_idx( pl_i2s->in.ihead, pl_i2s->in.slots);
				
				if (info.fragments <= 0)
					info.fragments = pl_i2s->in.slots + info.fragments;
			}

			info.fragstotal = pl_i2s->in.slots;
			info.fragsize = 1 << pl_i2s->in.order;
			info.bytes = info.fragments * info.fragsize;
			rc = copy_to_user((void *)arg, &info, sizeof(audio_buf_info));
			break;
		}
		
		case SNDCTL_DSP_SETFRAGMENT:
		{
			get_user_ret(val, (int *)arg, -EFAULT);
			if (file->f_mode & FMODE_WRITE) 
			{
				validate_state(pl_i2s->out, ST_OPEN, -EINVAL);
				rc = pl_dsp_setfragment(&pl_i2s->out, 0, ((unsigned int)val) >> 16, val & 0xffff );
			}
			
			if (file->f_mode & FMODE_READ) 
			{
				validate_state(pl_i2s->in, ST_OPEN, -EINVAL);
				rc = pl_dsp_setfragment(&pl_i2s->in, 1, ((unsigned int)val) >> 16, val & 0xffff );
			}
			
			break;
		}
		
		case SNDCTL_DSP_GETOPTR:
		{
			if (!(file->f_mode & FMODE_WRITE))
				return -EINVAL;
			pl_dsp_getoptr(&pl_i2s->out, &cinfo);
			rc = copy_to_user((void *)arg, &cinfo, sizeof(count_info));
			
			break;
		}
		
		case SNDCTL_DSP_GETIPTR:
		{
			if (!(file->f_mode & FMODE_READ))
				return -EINVAL;
			pl_dsp_getiptr(&pl_i2s->in, &cinfo);
			rc = copy_to_user((void *)arg, &cinfo, sizeof(count_info));
			
			break;
		}
		
		case SNDCTL_DSP_NONBLOCK:
		{
			file->f_flags |= O_NONBLOCK;
			return 0;
		}

		case SNDCTL_DSP_GETCAPS:
		{
			rc = put_user(DSP_CAP_DUPLEX | DSP_CAP_REALTIME | DSP_CAP_TRIGGER, (int *) arg);
			break;
		}
		
		case SNDCTL_DSP_SETTRIGGER:
		{
			get_user_ret(val, (int *)arg, -EFAULT);
			if (file->f_mode & FMODE_WRITE) 
			{
				validate_state(pl_i2s->out, ST_OPEN, -EINVAL);

				if (val & PCM_ENABLE_OUTPUT)
					start_issue(&pl_i2s->out);
				else
					stop_issue(&pl_i2s->out);

			}
			
			if (file->f_mode & FMODE_READ)
			{
				validate_state(pl_i2s->in, ST_OPEN, -EINVAL);

				if (val & PCM_ENABLE_INPUT)
					start_issue(&pl_i2s->in);
				else
					stop_issue(&pl_i2s->in);
			}
			
			break;
		}
		
		case SNDCTL_DSP_GETTRIGGER:
		{
			if (file->f_mode & FMODE_WRITE) 
			{
				if (pl_i2s->out.start)
					val |= PCM_ENABLE_OUTPUT;
			}
			
			if (file->f_mode & FMODE_READ) 
			{
				if (pl_i2s->in.start)
					val |= PCM_ENABLE_INPUT;
			}
			
			break;
		}

		case SNDCTL_DSP_GETBLKSIZE:
		{
			if (file->f_mode & FMODE_WRITE)
				val = 1 << pl_i2s->out.order;
			else if (file->f_mode & FMODE_READ)
				val = 1 << pl_i2s->in.order;
			
			rc = put_user(val, (int*)arg);
			
			break;
		}

		case SOUND_PCM_READ_RATE:
		{
			if (file->f_mode & FMODE_WRITE)
				val = pl_dsp_get_samplerate(pl_i2s, DIR_OUT);
			else if (file->f_mode & FMODE_READ)
				val = pl_dsp_get_samplerate(pl_i2s, DIR_IN);
			
			rc = put_user(val, (int *)arg);
			
			break;
		}

		case SOUND_PCM_READ_CHANNELS:
		{
			if (file->f_mode & FMODE_WRITE)
				val = pl_dsp_get_channels(DIR_OUT);
			else if (file->f_mode & FMODE_READ)
				val = pl_dsp_get_channels(DIR_IN);

			rc = put_user(val, (int *)arg);
			
			break;
		}
		
		case SOUND_PCM_READ_BITS:
		{
			if (file->f_mode & FMODE_WRITE)
				val = pl_dsp_get_format(DIR_OUT);
			else if (file->f_mode & FMODE_READ)
				val = pl_dsp_get_format(DIR_IN);

			rc = put_user(val, (int *)arg);
			
			break;
		}

		case SNDCTL_DSP_GETODELAY:
		{
			if (file->f_mode & FMODE_WRITE)
				val = pl_dsp_getodelay(&pl_i2s->out);
			else
				val = 0;


			rc = put_user(val, (int *)arg);
			
			break;
		}

		case SNDCTL_DSP_GETIDELAY:
		{
			if (file->f_mode & FMODE_READ)
				val = pl_dsp_getidelay(&pl_i2s->in);
			else
				val = 0;

			rc = put_user(val, (int *)arg);
			
			break;
		}

		case PL_SNDCTL_DSP_SYMALR_REQ :
		{
			if (file->f_mode & FMODE_WRITE) 
			{
				PLSymbolAlarm SymbolAlarm;
				int ord;

				validate_state(pl_i2s->out, ST_OPEN, -EINVAL);
				rc = copy_from_user(&SymbolAlarm, (void *)arg, sizeof(PLSymbolAlarm));
				if (rc < 0)
					return -EFAULT;
				
				if (SymbolAlarm.nPID == 0)
					SymbolAlarm.nPID = current->pid;
				
				if (SymbolAlarm.nMaxBufferSize > 0 && SymbolAlarm.nMaxBufferAmount > 0) 
				{
					ord = get_order_by_size(SymbolAlarm.nMaxBufferSize);
					rc = pl_dsp_setfragment(&pl_i2s->out, 0, SymbolAlarm.nMaxBufferAmount, ord);
					if (rc < 0)
						break;
				}
				SymbolAlarm.nMaxBufferAmount = pl_i2s->out.slots;
				SymbolAlarm.nMaxBufferSize = 1 << pl_i2s->out.order;
				pl_i2s->out.threshold = (SymbolAlarm.nFrequency > SymbolAlarm.nMaxBufferSize) ?
				SymbolAlarm.nMaxBufferSize : SymbolAlarm.nFrequency;
				pl_i2s->out.alarmpid = SymbolAlarm.nPID;
				rc = copy_to_user((void *)arg, &SymbolAlarm, sizeof(PLSymbolAlarm));
			} 
			else
				return -EINVAL;
			
			break;
		}

		case PL_SNDCTL_DSP_SYMALR_FREQUENCY:
		{
			if (file->f_mode & FMODE_WRITE)
			{
				get_user_ret(val, (int *)arg, -EFAULT);
				if (val > (1 << pl_i2s->out.order) || val < 0)
					return -EINVAL;
				
				pl_i2s->out.threshold = val;
				if (pl_i2s->out.alarmpid == 0)
					pl_i2s->out.alarmpid = current->pid;
			} else
				return -EINVAL;
			
			break;
		}
		
		case PL_SNDCTL_DSP_SYMALR_ACTIVE :
		{
			if (file->f_mode & FMODE_WRITE) 
			{
				get_user_ret(val, (int *)arg, -EFAULT);
				
				switch(val) 
				{
					case PL_SYMBOL_ALARM_QUERY:
						val = (pl_i2s->out.alarm) ? PL_SYMBOL_ALARM_ENABLE : PL_SYMBOL_ALARM_DISABLE;
						rc = put_user(val, (int *)arg);
						break;

					case PL_SYMBOL_ALARM_DISABLE:
						pl_i2s->out.alarm = 0;
						break;
					
					case PL_SYMBOL_ALARM_ENABLE:
						pl_i2s->out.alarm = 1;
						break;
						
					default:
						rc = -EINVAL;
						break;
				}
			} 
			else
				return -EINVAL;
			
			break;
		}
		
		case PL_SNDCTL_DSP_SYMALR_REL :
		{
			if (file->f_mode & FMODE_WRITE) 
			{
				pl_i2s->out.threshold = (1 << pl_i2s->out.order);
				pl_i2s->out.alarm = 0;
			} 
			else
				return -EINVAL;
			
			break;
		}
		
		default:
			printk("unhandled ioctl, cmd=%u arg=%ld\n", cmd, arg);
			rc = -EINVAL;
			
			break;
	}
	
	return rc;
}



static int pl_dsp_mmap(struct file *file, struct vm_area_struct *pvma)
{
	return -EINVAL;
}


static struct file_operations pl_dsp_ops =
{
	owner:      THIS_MODULE,
	open:       pl_dsp_open,
	release:    pl_dsp_release,
	read:       pl_dsp_read,
	write:      pl_dsp_write,
	ioctl:      pl_dsp_ioctl,
	poll:       pl_dsp_poll,
	llseek:     no_llseek,
	mmap:       pl_dsp_mmap,
};



/******************************************************************************
 *  Mixer functions
 ******************************************************************************/
static int pl_mixer_open(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	if (pl_i2s_ins.dev_mixer != minor)
		return -ENODEV;

	file->private_data = &pl_i2s_ins;
	return 0;
}

static int pl_mixer_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int reg=0, val=0, ccode=0;

	if (_IOC_TYPE(cmd) == 'A' && _SIOC_SIZE(cmd) == sizeof(int))
	{
		reg = _IOC_NR(cmd);

		if (_SIOC_DIR(cmd) == _SIOC_READ) 
		{
			AUDIO_DBG("Cannot read !!!\n");
			return -EFAULT;
		}
		else 
		{
			if (get_user(val, (int *)arg))
				return -EFAULT;
			ccode = aic23_write_value(reg,val);
			AUDIO_DBG("ccode = %d\n",ccode);
			return 0;
		}
	}
	
	return 0;

}

static int pl_mixer_release(struct inode *inode, struct file *file)
{
	return 0;
}


static struct file_operations pl_mixer_ops =
{
	owner:      THIS_MODULE,
	open:       pl_mixer_open,
	llseek:     no_llseek,
	ioctl:      pl_mixer_ioctl,
	release:    pl_mixer_release,
};



/******************************************************************************
 *  audio proc information
 ******************************************************************************/
static int pl_i2s_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	pl_i2s_t *pl_i2s = data;
	struct pl_i2s_info *out_info = &pl_i2s->out.info;
	struct pl_i2s_info *in_info = &pl_i2s->in.info;
	len += sprintf(page+len, "Prolific i2s driver info\n");
	len += sprintf(page+len, "63 Audio Module\n");
	len += sprintf(page+len, "Downstream:\n");
	len += sprintf(page+len, "  fragments: %d\n", out_info->slots);
	len += sprintf(page+len, "  underrun:  %d\n", out_info->runs);
	len += sprintf(page+len, "  jam:       %d\n", out_info->jam);
	len += sprintf(page+len, "  losts:     %d\n", out_info->losts);
	len += sprintf(page+len, "  restores:  %d\n", out_info->restores);
	len += sprintf(page+len, "  flags:     %s%s%s%s\n",
	    (out_info->flags & HW_FLAG_EMP) ? "EMP " : "",
	    (out_info->flags & HW_FLAG_FUL) ? "FUL " : "",
	    (out_info->flags & HW_FLAG_OVF) ? "OVF " : "",
	    (out_info->flags & HW_FLAG_UNF) ? "UNF " : "");
	len += sprintf(page+len, "Upstream:\n");
	len += sprintf(page+len, "  fragments: %d\n", in_info->slots);
	len += sprintf(page+len, "  overrun:   %d\n", in_info->runs);
	len += sprintf(page+len, "  jam:       %d\n", in_info->jam);
	len += sprintf(page+len, "  losts:     %d\n", in_info->losts);
	len += sprintf(page+len, "  restores:  %d\n", in_info->restores);
	len += sprintf(page+len, "  flags:     %s%s%s%s\n",
	    (in_info->flags & HW_FLAG_EMP) ? "EMP " : "",
	    (in_info->flags & HW_FLAG_FUL) ? "FUL " : "",
	    (in_info->flags & HW_FLAG_OVF) ? "OVF " : "",
	    (in_info->flags & HW_FLAG_UNF) ? "UNF " : "");

	return len;
}


/******************************************************************************
 *  Register and unregister device functions
 ******************************************************************************/


static void __init pl_i2s_init_proc(void)
{
	struct proc_dir_entry *audio_proc;

	audio_proc = proc_mkdir("audio", 0);
	if (audio_proc == NULL)
		goto EXIT;

	audio_proc->owner = THIS_MODULE;
	create_proc_read_entry("info", 0, audio_proc, pl_i2s_read_proc, &pl_i2s_ins);

EXIT:
	return;
}

int aic23_write_value( int reg, u16 val )
{
	int i;
	aic23_client->addr = 0x1a;
	struct i2c_client *client = aic23_client;
	
	if ((reg < 0 || reg > 9) && (reg != 15)) 
	{
		AUDIO_DBG("Invalid register R%d\n", reg);
		return -1;
	}

	for (i = 0; i < 3; i++) 
	{
		if (i2c_smbus_write_byte_data(client, (reg << 1) |
					(val >> 8), val & 0xff) == 0) 
		{
			return 0;
		}
	}
	AUDIO_DBG("I2C: cannot write %03x to register R%d\n", val, reg);
	return -1;
}

int aic23_detect_client(struct i2c_adapter *adapter, int address, unsigned short flags, int kind)
{
	int err = 0;
	const char *client_name = "TLV320AIC23 Audio Codec";
 
	if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_WORD_DATA |
				     I2C_FUNC_SMBUS_WRITE_BYTE)) 
	{
		printk(KERN_WARNING "%s functinality check failed\n",
		       client_name);
		return err;
	}
 
	if (!(aic23_client = kmalloc(sizeof(struct i2c_client),
				   GFP_KERNEL)))
	{
		err = -ENOMEM;
		printk(KERN_WARNING "Couldn't allocate memory for %s\n",
		       client_name);
		return err;
	}
  
	memset(aic23_client, 0x00, sizeof(struct i2c_client));
	aic23_client->addr = address;
	AUDIO_DBG("addr = %d\n", aic23_client->addr );
	aic23_client->adapter = adapter;
	aic23_client->driver = &aic23_driver;
	aic23_client->flags = 0;
	strcpy(aic23_client->name, client_name);
                                                     
	if ((err = i2c_attach_client(aic23_client))) 
	{
		printk(KERN_WARNING "Couldn't attach %s\n", client_name);
		kfree(aic23_client);
		return err;
	}
	                                                     
	return 0;
}

static int aic23_detach_client(struct i2c_client *client)
{
	int err;

	if ((err = i2c_detach_client(client))) 
	{
		printk("aic23.o: Client deregistration failed, \
		       client not detached.\n");
		return err;
	}
	kfree(client);
	return 0;
}

static int aic23_attach_adapter(struct i2c_adapter *adapter)
{
	int res;
	if (!adapter)
	{
		                                                     
		return -NODEV;
	}
	                                                     
	res = i2c_detect(adapter, &addr_data, aic23_detect_client);
	AUDIO_DBG("res = %d\n", res);
	                                                     
	return res;
}


static struct i2c_driver aic23_driver = {
	.owner			= THIS_MODULE,
	.name			= "TLV320AIC23 codec",	//	max. 40 characters limited
	.id				= I2C_DRIVERID_TLV320AIC23,			//	usually from i2c-id.h
	.flags			= I2C_DF_NOTIFY,				//	always set this value
 	.attach_adapter	= aic23_attach_adapter,
	.detach_client	= aic23_detach_client,
};

// input , level 0~20, level 0: mute
static void update_volume_left(int level)
{
	u16 val = 0;
	if (level >= AUDIO_VOLUME_TABLEN)
		return ;
		
	val =  LRSH_ON|  LZC_ON| HPVOLUME_TAB[level];
	aic23_write_value(LEFT_CHANNEL_VOLUME_ADDR, val);
	aic23_info_l.volume_reg_left = level;
}

// input , level 0~20, level 0: mute
static void update_volume_right(int level)
{
	u16 val = 0;
	if (level > AUDIO_VOLUME_TABLEN)
		return ;
		
	val =  LRSH_ON|  LZC_ON| HPVOLUME_TAB[level];
	aic23_write_value(RIGHT_CHANNEL_VOLUME_ADDR, val);
	aic23_info_l.volume_reg_right = level;
}

#if 0
static void update_input_gain_left(int gain)
{
	u16 val = 0;
	val = ((gain * INPUT_VOLUME_RANGE) / 100) + INPUT_VOLUME_MIN;
	aic23_write_value(LEFT_LINE_VOLUME_ADDR, val);
	aic23_info_l.input_gain_reg_left = gain;
}

static void update_input_gain_right(int gain)
{
	u16 val = 0;
	val = ((gain * INPUT_VOLUME_RANGE) / 100) + INPUT_VOLUME_MIN;
	aic23_write_value(RIGHT_LINE_VOLUME_ADDR, val);
	aic23_info_l.input_gain_reg_right = gain;
}


static void aic23_audio_shutdown(void)
{
	/* Let's mute the codec before powering off to prevent
	* glitch in the sound
	*/
	aic23_write_value(LEFT_CHANNEL_VOLUME_ADDR, LHV_MIN);
	aic23_write_value(RIGHT_CHANNEL_VOLUME_ADDR, LHV_MIN);
	aic23_power_down();
}

static int aic23_audio_suspend(void)
{
	/* Let's mute the codec before powering off to prevent
	 * glitch in the sound
	 */
	aic23_write_value(LEFT_CHANNEL_VOLUME_ADDR, LHV_MIN);
	aic23_write_value(RIGHT_CHANNEL_VOLUME_ADDR, LHV_MIN);
	aic23_power_down();

	return 0;
}

static int aic23_audio_resume(void)
{
	aic23_power_up();

	return 0;
}
#endif

#if 0
static void set_mic(int mic_en)
{
	u16 dg_ctrl;

	if (mic_en) 
	{
		aic23_info_l.power = 0x00;//OSC_OFF | LINE_OFF;
		dg_ctrl = ADCHP_ON;
		aic23_info_l.mask &= ~MICM_MUTED;
		aic23_info_l.mask |= MICB_20DB;
	} 
	else 
	{
		aic23_info_l.power =
			OSC_OFF | ADC_OFF | MIC_OFF | LINE_OFF;
		dg_ctrl = 0x00;
		aic23_info_l.mask =
			DAC_SELECTED | INSEL_MIC | MICM_MUTED;
	}
	aic23_write_value(POWER_DOWN_CONTROL_ADDR,
				aic23_info_l.power);
	aic23_write_value(DIGITAL_AUDIO_CONTROL_ADDR, dg_ctrl);
	aic23_write_value(ANALOG_AUDIO_CONTROL_ADDR,  aic23_info_l.mask);
	aic23_info_l.mic_enable = mic_en;

	AUDIO_DBG(KERN_INFO "aic23 mic state: %i\n", mic_en);
}
#endif

static void aic23_init_power_up(void)
{

	
	aic23_write_value(RESET_CONTROL_ADDR, 0x00);      /*Reset the codec*/
	
	aic23_write_value(POWER_DOWN_CONTROL_ADDR, 0x00); /*Enale the device*/
	
	aic23_write_value(ANALOG_AUDIO_CONTROL_ADDR, DAC_SELECTED|INSEL_MIC|MICB_20DB);
	
	aic23_write_value(DIGITAL_AUDIO_CONTROL_ADDR, ADCHP_ON);
	
	aic23_write_value(DIGITAL_AUDIO_FORMAT_ADDR, LRP_ON | FOR_I2S);        /*I2S mode*/
	
	
	aic23_write_value(SAMPLE_RATE_CONTROL_ADDR, SR1SR|SR0SR|USB_CLK_ON);   /*sample rate : 8K*/
	
	aic23_write_value(DIGITAL_INTERFACE_ACT_ADDR, ACT_ON);                 /*activate */
	
	aic23_write_value(LEFT_CHANNEL_VOLUME_ADDR, LHV_MAX);
	aic23_write_value(RIGHT_CHANNEL_VOLUME_ADDR, LHV_MAX);
	
	aic23_info_l.power_down = 0;

	AUDIO_DBG(KERN_INFO "aic23_init_power() done\n");

#if 0
	if (aic23_info_l.initialized == 0) 
	{
		AUDIO_LINE;
		aic23_write_value(LEFT_CHANNEL_VOLUME_ADDR, LHV_MIN);
		aic23_write_value(RIGHT_CHANNEL_VOLUME_ADDR, LHV_MIN);
	}
	else 
	{
		AUDIO_LINE;
		update_volume_left(aic23_info_l.volume_reg_left);
		update_volume_right(aic23_info_l.volume_reg_right);
	}
	aic23_write_value(LEFT_CHANNEL_VOLUME_ADDR, LHV_MAX);
	aic23_write_value(RIGHT_CHANNEL_VOLUME_ADDR, LHV_MAX);
	
	aic23_info_l.mask = DAC_SELECTED | INSEL_MIC | MICM_MUTED ;
	aic23_write_value(ANALOG_AUDIO_CONTROL_ADDR,
			        aic23_info_l.mask);
	aic23_write_value(DIGITAL_AUDIO_CONTROL_ADDR, 0x00);
	aic23_write_value(DIGITAL_AUDIO_FORMAT_ADDR, LRP_ON | FOR_I2S);
	aic23_write_value(SAMPLE_RATE_CONTROL_ADDR, USB_CLK_ON);
	aic23_write_value(DIGITAL_INTERFACE_ACT_ADDR, ACT_ON);
	aic23_info_l.power = OSC_OFF | ADC_OFF | MIC_OFF | LINE_OFF;
	aic23_write_value(POWER_DOWN_CONTROL_ADDR,
			        aic23_info_l.power);

	/* enable mic input */
	if (aic23_info_l.mic_enable)
		set_mic(aic23_info_l.mic_enable);
	
#ifdef _AIC23HW_BUG_
	aic23_write_value(RESET_CONTROL_ADDR, 	0x00);
	aic23_write_value(POWER_DOWN_CONTROL_ADDR, 0x00);
	aic23_write_value(ANALOG_AUDIO_CONTROL_ADDR, 0x15);
	aic23_write_value(DIGITAL_AUDIO_CONTROL_ADDR, 0x01);
	aic23_write_value(DIGITAL_AUDIO_FORMAT_ADDR, 0x12);
	aic23_write_value(SAMPLE_RATE_CONTROL_ADDR, 0x0d);
	aic23_write_value(DIGITAL_INTERFACE_ACT_ADDR, 0x01);
	aic23_write_value(LEFT_CHANNEL_VOLUME_ADDR, 0x7f);
	aic23_write_value(RIGHT_CHANNEL_VOLUME_ADDR, 0x7f);
#endif
#endif

}

void aic23_power_up(void)
{
	if (aic23_info_l.power_down)
	{
		AUDIO_DBG("aic23 powering up\n");
		aic23_init_power_up();
	}
}

void aic23_init_power_down(void)
{
	/* Let's mute the codec before powering off to prevent
 		* glitch in the sound
 	*/
	aic23_write_value(LEFT_CHANNEL_VOLUME_ADDR, LHV_MIN);
	aic23_write_value(RIGHT_CHANNEL_VOLUME_ADDR, LHV_MIN);
	aic23_write_value(POWER_DOWN_CONTROL_ADDR, 0xff);
	aic23_info_l.power_down = 1;
	AUDIO_DBG("aic23 powering down\n");
}

void aic23_power_down(void)
{
	if (!aic23_info_l.power_down) 
	{
		aic23_init_power_down();
	}
}

static int __init pl_i2s_init(void)
{

	selftest =  0;

	memset(&pl_i2s_ins, 0, sizeof(pl_i2s_t));
	memset(&aic23_info_l, 0, sizeof(aic23_info_t));
	                                                     
	/*  audio controller reset */
	writeb(0, PL_AUDIO_CMD_COMMAND_1);
	writeb(COMMAND_1_B_RSTN, PL_AUDIO_CMD_COMMAND_1);
                                                     
	/* reset  function register */
	writel(0, PL_AUDIO_CMD_FUNCTION_0);
	writeb(FUNCTION_1_VRA, PL_AUDIO_CMD_FUNCTION_1);
                                                     
	/* set internal clock and divider enable */
	writel(0x04, PL_AUDIO_CMD_CLOCK);
	writeb(0x02, PL_CLK_AC97);
                                                     
	writeb(0, PL_AUDIO_CMD_GPIO_C_ENB);     /* select pin as I2S function */
	writeb(0x0f, PL_AUDIO_CMD_GPIO_C_OE);   /* for I2S should be 01111 internal clock */
                                                                                                          
	/* Init aic23 by I2C command */
	if (i2c_add_driver(&aic23_driver)) 
	{
		printk("aic23 i2c: Driver registration failed, module not inserted.\n");
		selftest = -ENODEV;
		goto EXIT_1;
	}
	if( !aic23_client)
	{
		printk("aic23 i2c: Driver registration failed, module not inserted.\n");
		selftest = -ENODEV;
		goto EXIT_1;
	}

	aic23_init_power_down();

	writel(0x04, PL_AUDIO_CMD_CLOCK);
	
	writeb(COMMAND_0_I2S, PL_AUDIO_CMD_COMMAND_0);	  /* set i2s mode */                                                   
	writeb(I2S_WSO_MODE_OUTPUT| I2S_WSI_MODE_OUTPUT, PL_AUDIO_CMD_FUNCTION_2);   /* enable I2S up/down format */	                                                     
	writel(0, PL_AUDIO_CMD_FUNCTION_0);

	/*
	* Register a dsp device
	*/

	pl_i2s_ins.dev_mixer = register_sound_mixer(&pl_mixer_ops, -1);
	if (pl_i2s_ins.dev_mixer < 0)
	{
		printk("unable to register I2S mixer, aborting\n");
		selftest = pl_i2s_ins.dev_mixer;
		goto EXIT;
	}
                                          
	pl_i2s_ins.dspo = register_sound_dsp(&pl_dsp_ops, -1);
	if (pl_i2s_ins.dspo < 0)
	{
		printk("unable to register dsp output device, aborting\n");
		selftest = pl_i2s_ins.dspo;
		goto EXIT;
	}
                                                     
	pl_i2s_ins.dspi = register_sound_dsp(&pl_dsp_ops, -1);
	if (pl_i2s_ins.dspi < 0) 
	{
		printk("unable to register dsp output device, aborting\n");
		selftest = pl_i2s_ins.dspi;
		goto EXIT;
	}
                                           
	/* initial out/in buffer */
	pl_dsp_setfragment(&pl_i2s_ins.out, 0, DEFAULT_OUT_BUF_SLOTS, DEFAULT_OUT_BUF_SLOT_SIZE_ORDER);
	init_waitqueue_head(&pl_i2s_ins.out.wait);
	init_waitqueue_head(&pl_i2s_ins.out.drain);
	atomic_set(&pl_i2s_ins.out.issues, 0);
	pl_i2s_ins.out.issue = issue_out;

	pl_dsp_setfragment(&pl_i2s_ins.in, 1, DEFAULT_IN_BUF_SLOTS, DEFAULT_IN_BUF_SLOT_SIZE_ORDER);
	init_waitqueue_head(&pl_i2s_ins.in.wait);
	init_waitqueue_head(&pl_i2s_ins.in.drain);
	atomic_set(&pl_i2s_ins.in.issues, 0);
	pl_i2s_ins.in.issue = issue_in;
                                                     
	/* initial default sample rate */
	pl_i2s_ins.osrate = 48000;
	pl_i2s_ins.israte = 8000;
	pl_i2s_ins.micboost = 0;
	pl_i2s_ins.modcnt = 0;
	pl_i2s_ins.power = 0;
                                                     
	pl_i2s_ins.mixer_state[0] = (12 << 8) | 12;
	pl_i2s_ins.mixer_state[1] = (50 << 8) | 50;
	pl_i2s_ins.mixer_state[2] = (36<<8) | 36;

	/* init dma channel */
	selftest = request_dma(DMA_PL_AC97_OUT, "I2S DS");
	if (selftest < 0)
		goto EXIT;
                                                     
	set_dma_mde_base(DMA_PL_AC97_OUT, (const char *)pl_i2s_ins.out.mde);
	set_dma_mode(DMA_PL_AC97_OUT, DMA_MODE_WRITE);
	set_dma_burst_length(DMA_PL_AC97_OUT, 0);

	selftest = request_dma(DMA_PL_AC97_IN, "I2S US");
	if (selftest < 0)
		goto EXIT;

	set_dma_mde_base(DMA_PL_AC97_IN, (const char *)pl_i2s_ins.in.mde);
	set_dma_mode(DMA_PL_AC97_IN, DMA_MODE_READ);
	set_dma_burst_length(DMA_PL_AC97_IN, 0);

	/* init irq handler */
	selftest = request_irq(IRQ_PL_AC970, pl_i2s_handler, 0, "Audio I2S", &pl_i2s_ins);
	if (selftest < 0)
		goto EXIT;

	/* i2s controller enable interrupt */
	writeb(COMMAND_1_B_RSTN | COMMAND_1_INTE_DNDA | COMMAND_1_INTE_UPDA, PL_AUDIO_CMD_COMMAND_1);

	/* create proc entry */
	pl_i2s_init_proc();
	
	printk("AIC23 Driver(Ver: %s, Release-Date: %s)\n", TLV320AIC23_VERSION, TLV320AIC23_DATE);
	
	return 0;
EXIT:
	if (pl_i2s_ins.dspi > 0)
		unregister_sound_dsp(pl_i2s_ins.dspi);

	if (pl_i2s_ins.dspo > 0)
		unregister_sound_dsp(pl_i2s_ins.dspo);

	if (pl_i2s_ins.dev_mixer > 0)
		unregister_sound_dsp(pl_i2s_ins.dev_mixer);
	
	if( aic23_client)
		i2c_del_driver(&aic23_driver);
EXIT_1:

	return selftest;
}

module_init(pl_i2s_init);

#ifdef CONFIG_MODULES
static void __exit pl_i2s_clean(void)
{
	printk(KERN_INFO "pl_i2s: unloading\n");
}
module_exit(pl_i2s_clean);
#endif


	/**
	* Grammar:
	*    pl_audio=[outbuf(size_order,number)][,inbuf(size_order,number)]
	*    the size is a fragment order size, it can be 256, 512, 1024, 2048, 4096(defult), 8192 and 16384
	*    the number is a maxiumn fragments, it can be 4 ~ 64
	*/
static int __init pl_i2s_setup(char *options)
{
	int size;
	if (strncmp(options, "inbuf(",6) == 0)
	{
		options += 6;
		size = simple_strtoul(options, &options, 0);
		DEFAULT_OUT_BUF_SLOT_SIZE_ORDER = get_order_by_size(size);
		if (*options != ',')
			goto EXIT;
		options++;
		DEFAULT_OUT_BUF_SLOTS = simple_strtoul(options, &options, 0);
		if (*options != ')')
			goto EXIT;
		options++;

		if (*options == ',')
			options++;
	}

	if (strncmp(options, "outbuf(", 7) == 0) 
	{
		options += 7;
		size = simple_strtoul(options, &options, 0);
		DEFAULT_IN_BUF_SLOT_SIZE_ORDER = get_order_by_size(size);
		if (*options != ',')
			goto EXIT;
		options++;
		DEFAULT_IN_BUF_SLOTS = simple_strtoul(options, &options, 0);
		if (*options != ')')
			goto EXIT;
		options++;
	}
	
EXIT:

	if (DEFAULT_OUT_BUF_SLOT_SIZE_ORDER < 8)
		DEFAULT_OUT_BUF_SLOT_SIZE_ORDER = 8;
	else if (DEFAULT_OUT_BUF_SLOT_SIZE_ORDER > 15)
		DEFAULT_OUT_BUF_SLOT_SIZE_ORDER = 15;

	if (DEFAULT_IN_BUF_SLOT_SIZE_ORDER < 8)
		DEFAULT_IN_BUF_SLOT_SIZE_ORDER = 8;
	else if (DEFAULT_IN_BUF_SLOT_SIZE_ORDER > 15)
		DEFAULT_IN_BUF_SLOT_SIZE_ORDER = 15;

	if (DEFAULT_OUT_BUF_SLOTS > 64)
		DEFAULT_OUT_BUF_SLOTS = 64;
	if (DEFAULT_IN_BUF_SLOTS > 64)
		DEFAULT_IN_BUF_SLOTS = 64;

	if (DEFAULT_OUT_BUF_SLOTS <= 0)
		DEFAULT_OUT_BUF_SLOTS = 4;

	if (DEFAULT_IN_BUF_SLOTS <= 0)
		DEFAULT_IN_BUF_SLOTS = 4;


	return 1;
}

__setup("pl_audio=", pl_i2s_setup);

MODULE_AUTHOR("www.sercomm.com");
MODULE_DESCRIPTION("AIC23 i2s audio driver");

