#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <video/plmedia.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <asm/uaccess.h>

#include "sony098.h"

#if 0
#define X_PRINTK 	printk
#else
#define X_PRINTK(...)
#endif

#define		LM_DATE		"20070727"
#define		LM_VERSION	"1.0.1.1"

MODULE_LICENSE("Sercomm");


PSONY098DATA	g_pSONY098Data;
PSENSOR			g_pPLSensor;


static const char SONY098_SHORT_NAME[]="SONY098";
static const char SONY098_FULL_NAME[]="SONY ICX098BQ Sensor";
static const char SONY098_VERSION[]="1.0.1.1";

static int SensorOpen(struct inode *pinode, struct file *pfile)
{
	return 0;
}

static int SeneorRelese(struct inode *pinode, struct file *pfile)
{
	return 0;
}

static int SensorIOCTL(struct inode *pinode, struct file *pfile, unsigned int cmd, unsigned long arg)
{
	REGISTERINFO		reg;

	switch(cmd) {
		case PLSENSOR_WRITE_REGISTER:
			copy_from_user((void*)&reg , (void*)arg, sizeof(REGISTERINFO));
			//OV9653WriteByte(pClient, reg.uRegister, reg.uValue);
			break;
		case PLSENSOR_READ_REGISTER:
			copy_from_user((void*)&reg , (void*)arg, sizeof(REGISTERINFO));
			//reg.uValue=OV9653ReadByte(pClient, reg.uRegister);
			//copy_to_user((void*)arg, (void*)&reg, sizeof(REGISTERINFO));
			break;
	}

	return 0;
}

static struct file_operations SensorFOPS = 
{
	owner:		THIS_MODULE,
	open:		SensorOpen,
	release:	SeneorRelese,
	ioctl:		SensorIOCTL,
};


static void SONY098Restart(PSENSOR pSensor)
{
	/* restart the hardware */
	X_PRINTK("SONY098Restart()\n");

	return;
}

static int SONY098SetFormat(PSENSOR pSensor)
{
	/* set the video format to the ccd */
	pSensor->Format.nInterlaced=0;
	pSensor->Format.nEvenFieldFirst=0;

	//if (pSensor->Format.nCropWidth <= 640 && pSensor->Format.nCropHeight <= 512)
	printk("<1>Switch to context A\n");

	pSensor->Format.nFPS=30;
	pSensor->VideoStandard.FrameRate.uNumerator=1;
	pSensor->VideoStandard.FrameRate.uDenominator=30;
	pSensor->VideoStandard.uFrameLines=525;
	pSensor->VideoStandard.uFields=60;
	pSensor->VideoStandard.uColourStandard=VSTD_COLOR_STD_NTSC;
	pSensor->VideoStandard.ColourStandardData.NTSC.uColourSubcarrier=VSTD_COLOR_SUBC_NTSC;

	return 0;
}

static int SONY098GetFormat(PSENSOR pSensor)
{
	/* get the video format from the ccd */
	printk("SONY098GetFormat()\n");

	printk("nCropWidth: %d\tnCropHeight: %d\n", pSensor->Format.nCropWidth, pSensor->Format.nCropHeight);

	printk("\n");

	return 0;
}

static int SONY098QueryFormat(PSENSOR pSensor, int w, int h, int fps)
{
	printk("w: %d, h: %d, fps: %d\n", w, h, fps);

	if((w>640 && w<=1280) && (h>512 && h<=1024) && (fps<=15))
		return 0;
	
	if((w>0 && w<=640) && (h>0 && h<=512) && (fps<=30))
	return 0;	
		
	return -EFAULT;;
}



static int SONY098SetSync(PSENSOR pSensor)
{
	pSensor->SyncInfo.byHSyncState=0;
	pSensor->SyncInfo.byVSyncState=1;

	//printk("pSensor->SyncInfo.byITUBT656: %d\n", pSensor->SyncInfo.byITUBT656);
	return 0;
}

static int SONY098GetSync(PSENSOR pSensor)
{
	/** get the video sync from the ccd, return nothing so far. **/
	X_PRINTK("SONY098GetSync()\n");
	
	return 0;	
}

static int SONY098SetInput(PSENSOR pSensor)
{
	printk("SONY098SetInput\n");
	pSensor->Format.nInterlaced=0;
	return 0;
}

static int SONY098GetInput(PSENSOR pSensor)
{
	printk("SONY098GetInput\n");
	return 0;
}

static int SONY098SetVSTD(PSENSOR pSensor)
{
	printk("SONY098SetVSTD\n");
	//SetVideoStandard(pSensor);
	return 0;
}

static int SONY098GetVSTD(PSENSOR pSensor)
{
	printk("SONY098GetVSTD\n");
	//SetVideoStandard(pSensor);
	return 0;
}

static int SONY098ShowStatus(PSENSOR pSensor)
{
	return 0;
}

static void SONY098InitClient(PSONY098DATA data)
{
	data->pSensor->nType		= TYPE_SENSOR;	
    data->pSensor->szShortName	= SONY098_SHORT_NAME;
    data->pSensor->szFullName	= SONY098_FULL_NAME;
    data->pSensor->szVersion	= SONY098_VERSION;
	data->pSensor->PrivateData	= data;				/** it's free to use, the driver does not use it so far **/
	data->pSensor->Restart		= SONY098Restart;
	data->pSensor->QueryFormat	= SONY098QueryFormat;
	data->pSensor->SetFormat	= SONY098SetFormat;
	data->pSensor->GetFormat	= SONY098GetFormat;
	data->pSensor->SetSync		= SONY098SetSync;
	data->pSensor->GetSync		= SONY098GetSync;
	
	data->pSensor->SetInput		= SONY098SetInput;
	data->pSensor->GetInput		= SONY098GetInput;
	data->pSensor->SetVSTD		= SONY098SetVSTD;
	data->pSensor->GetVSTD		= SONY098GetVSTD;
	data->pSensor->ShowStatus	= SONY098ShowStatus;

	data->pSensor->SyncInfo.nHSyncStart		= 0;
	data->pSensor->SyncInfo.nVSyncStart		= 0;
	data->pSensor->SyncInfo.byITUBT656		= 1;    /* embedded sync */
	data->pSensor->SyncInfo.byHRefSelect	= 0;	
	data->pSensor->SyncInfo.nHSyncPhase		= 0;
	data->pSensor->SyncInfo.byHSyncState	= 0;
	data->pSensor->SyncInfo.byVSyncState	= 0;
	data->pSensor->SyncInfo.byValidEnable	= 0;
	data->pSensor->SyncInfo.byValidState	= 0;
	
	data->pSensor->Format.nCropWidth		= 640;
	data->pSensor->Format.nCropHeight		= 480;

	//data->pSensor->uVSTD						= VSTD_STD_NTSC_M;
	//SetVideoStandard(data->pSensor);

	data->pSensor->Format.nFPS					= 30;
	data->pSensor->Format.FPSEx.uNumerator		= 1;
	data->pSensor->Format.FPSEx.uDenominator	= 30;

	//printk("nNumberator: %d uDenominator: %d\n", data->pSensor->Format.FPSEx.uNumerator, data->pSensor->Format.FPSEx.uDenominator);
	
	// auto major and minor number
	data->DevFS = devfs_register (NULL, "sensor", DEVFS_FL_AUTO_DEVNUM, 0, 0, S_IFCHR | S_IWUSR | S_IRUSR, &SensorFOPS, NULL);	
	g_pPLSensor = data->pSensor;

	/* Reset HW & Set the register */
	X_PRINTK("Complete.\n");
}


static int __init sm_SONY098_init(void) {
	X_PRINTK("<1>plsensor.o  version %s (%s) (CCIR)\n", LM_VERSION, LM_DATE);

	usb_pwc_init1();

	if (!(g_pSONY098Data = kmalloc(sizeof(SONY098DATA)+sizeof(SENSOR), GFP_KERNEL))) 
	{
		return -1;
	}

	g_pPLSensor = g_pSONY098Data->pSensor =(PSENSOR)(g_pSONY098Data+1);

	
	SONY098InitClient(g_pSONY098Data);

	X_PRINTK("SyncInfo.byITUBT656: %d\n", g_pSONY098Data->pSensor->SyncInfo.byITUBT656);

	return 0;
}

static void __exit sm_SONY098_exit(void)
{
	X_PRINTK("<1>Remove Modules\n");
	usb_pwc_exit1();
	kfree(g_pSONY098Data);
}

MODULE_AUTHOR("Rondall Tsao");
MODULE_DESCRIPTION("SONY098BQ/TI 3in1/TP6842D Driver");

module_init(sm_SONY098_init);
module_exit(sm_SONY098_exit);

EXPORT_SYMBOL(g_pPLSensor);



