
/***************************************************************
 * (C) COPYRIGHT SerComm Corporation All Rights Reserved, 2008.
 *
 *      Use of this software is restricted to the terms and
 *      conditions of SerComm's software license agreement.
 *
 *                        www.sercomm.com
 ****************************************************************/


#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include "i2c-dev.h"

#include <video/plmedia.h>


static int ov_read_byte_data(int fd, int reg)
{
	int data = 0;
	write(fd, &reg, 1);
	read(fd, &data, 1);
	return data;

}

static int ov_write_byte_data(int fd, int reg, int val)
{
	return i2c_smbus_write_byte_data(fd, reg, val);
}
void usage(char *s)
{
	printf("Standard I2C device read/write\n");
	printf("Read:\t<%s> <addr> <reg>\n",s);
	printf("Write:\t<%s> <addr> <reg> <Write Byte>\n",s);
}
int main(int argc, char *argv[])
{
	int fd;
	int reg, val,addr;
	char fname[32];
	int i;
	long funcs;

	strncpy(fname, "/dev/i2c/0", sizeof(fname));

	fd = open(fname, O_RDWR);
	if (fd <= 0) 
	{
		printf("open device (%s) fail!\n", fname);
		return -1;
	}

	// check adapter functionality
	i = ioctl(fd, I2C_FUNCS, &funcs);
	if (i < 0) 
	{
		printf("ioctl I2C_FUNCS fail!\n");
		goto done;
	}
	// use FORCE so that we can write registers even when
	//	a driver is also running
	if (argc > 1)
	{
		addr = strtol(argv[1], NULL, 16);
	}
	else
	{
		usage(argv[0]);
		return -1;
	}
	i = ioctl(fd, I2C_SLAVE_FORCE, addr);
	if (i < 0) 
	{
		printf("ioctl I2C_SLAVE_FORCE fail!\n");
		goto done;
	}


	reg = 0x00;
	val = -1;

	if (argc > 2) 
	{
		reg = strtol(argv[2], NULL, 16);
	}
	else
	{
		usage(argv[0]);
		return -1;
	}
	
	if (argc > 3) 
	{
		val = strtol(argv[3], NULL, 16);
	}

	if (val < 0) 
	{
		val = ov_read_byte_data(fd, reg);
		printf("get reg(%.2X)=%.2X\n", reg, val);
	} 
	else 
	{
		if (ov_write_byte_data(fd, reg, val) < 0) 
		{
			printf("Set reg(%.4X) to val(%.4X) fail!\n", reg, val);
		}
		printf("set reg(%.2X)=%.2X\n", reg, val);
	}

done:
	close(fd);

	return 0;
}
