#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#include "ov7670.h"
#include "i2c-dev.h"

static inline int ov_read_byte_data(int fd, int reg)
{
	int data = 0;

	write(fd, &reg, 1);
	read(fd, &data, 1);

	return data;
}

typedef	struct st_reginfo {
	unsigned int	adr;
	unsigned int	val;
} reginfo;
reginfo OV7670_YUV_VGA_OptimalSet[] = 
{
	{OV7670REG_COM7,	0x80},
	// ;
	{OV7670REG_CLKRC,	0x01},
	{OV7670REG_TSLB,	0x04},
	{OV7670REG_COM7,	0x00},
	{OV7670REG_HSTART,	0x13},
	{OV7670REG_HSTOP,	0x01},
	{OV7670REG_HREF,	0xb6},
	{OV7670REG_VSTART,	0x02},
	{OV7670REG_VSTOP,	0x7a},
	{OV7670REG_VREF,	0x0a},
	{OV7670REG_COM3,	0x00},
	{OV7670REG_COM14,	0x00},
	{OV7670REG_SCALING_XSC,			0x3a},
	{OV7670REG_SCALING_YSC,			0x35},
	{OV7670REG_SCALING_DCWCTR,		0x11},
	{OV7670REG_SCALING_PCLK_DIV,	0xf0},
	{OV7670REG_SCALING_PCLK_DELAY,	0x02},
	// ;gamma  
	{OV7670REG_SLOP,	0x1d},
	{OV7670REG_GAM1,	0x06},
	{OV7670REG_GAM2,	0x0f},
	{OV7670REG_GAM3,	0x21},
	{OV7670REG_GAM4,	0x44},
	{OV7670REG_GAM5,	0x52},
	{OV7670REG_GAM6,	0x61},
	{OV7670REG_GAM7,	0x6e},
	{OV7670REG_GAM8,	0x79},
	{OV7670REG_GAM9,	0x84},
	{OV7670REG_GAM10,	0x8d},
	{OV7670REG_GAM11,	0x9d},
	{OV7670REG_GAM12,	0xab},
	{OV7670REG_GAM13,	0xc4},
	{OV7670REG_GAM14,	0xd9},
	{OV7670REG_GAM15,	0xea},
	// ;Based AEC/AGC
	{OV7670REG_COM8,	0xe0},
	{OV7670REG_GAIN,	0x00},
	{OV7670REG_AECH,	0x00},
	{OV7670REG_COM4,	0x40},
	{OV7670REG_COM9,	0x48},
	{OV7670REG_DB50MAX,	0x05},
	{OV7670REG_DB60MAX,	0x07},
	{OV7670REG_AEW,		0x95},
	{OV7670REG_AEB,		0x33},
	{OV7670REG_VPT,		0xe3},
	{OV7670REG_HAECC1,	0x80},
	{OV7670REG_HAECC2,	0x70},
	{0xa1,				0x03},	// ;cb white pixels correction enable.
	{OV7670REG_HAECC3,	0xdf},
	{OV7670REG_HAECC4,	0xdf},
	{OV7670REG_HAECC5,	0xf0},
	{OV7670REG_HAECC6,	0x90},
	{OV7670REG_HAECC7,	0x94},
	{OV7670REG_COM8,	0xe5},
	// ;
	{OV7670REG_COM5,	0x61},
	{OV7670REG_COM6,	0x4b},
	{0x16,				0x02},
	{OV7670REG_MVFP,	0x27},
	{OV7670REG_ADCCTR1,	0x02},
	{OV7670REG_ADCCTR2,	0x91},
	{0x29,				0x07},
	{OV7670REG_CHLF,	0x0b},
	{0x35,				0x0b},
	{OV7670REG_ADC,		0x1d},
	{OV7670REG_ACOM,	0x71},
	{OV7670REG_OFON,	0x2a},
	{OV7670REG_COM12,	0x78},
	{0x4d,				0x40},
	{0x4e,				0x20},
	{OV7670REG_GFIX,	0x00},
	{OV7670REG_DBLV,	0x4a},
	{OV7670REG_REG74,	0x19},
	{0x8d,				0x4f},
	{0x8e,				0x00},
	{0x8f,				0x00},
	{0x90,				0x00},
	{0x91,				0x00},
	{0x96,				0x00},
	{0x9a,				0x80},
	{0xb0,				0x84},
	{OV7670REG_ABLC1,	0x0c},
	{0xb2,				0x0e},
	{OV7670REG_THL_ST,	0x82},
	{0xb8,				0x0a},
//	{OV7670REG_ADVFL,	0x00},
//	{OV7670REG_ADVFH,	0x00},
	// ;AWB Control 
	{OV7670REG_AWBC1,	0x0a},
	{OV7670REG_AWBC2,	0xf0},
	{OV7670REG_AWBC3,	0x34},
	{OV7670REG_AWBC4,	0x58},
	{OV7670REG_AWBC5,	0x28},
	{OV7670REG_AWBC6,	0x3a},
	{0x59,				0x88},
	{0x5a,				0x88},
	{0x5b,				0x44},
	{0x5c,				0x67},
	{0x5d,				0x49},
	{0x5e,				0x0e},
	{OV7670REG_AWBCTR3,	0x0a},
	{OV7670REG_AWBCTR2,	0x55},
	{OV7670REG_AWBCTR1,	0x11},
	{OV7670REG_AWBCTR0,	0x9f},	// ;9e for advance AWB
	{OV7670REG_GGAIN,	0x40},
	{OV7670REG_BLUE,	0x40},
	{OV7670REG_RED,		0x40},
	{OV7670REG_COM8,	0xe7},
	// ;color matrix 
	{OV7670REG_MTX1,	0x88},
	{OV7670REG_MTX2,	0x8b},
	{OV7670REG_MTX3,	0x04},
	{OV7670REG_MTX4,	0x11},
	{OV7670REG_MTX5,	0x8c},
	{OV7670REG_MTX6,	0x9d},
	{OV7670REG_BRIGHT,	0x00},
	{OV7670REG_CONTRAS,	0x40},
	{OV7670REG_CONTRAS_CENTER,	0x80},
	{OV7670REG_MTXS,	0x9a},
	// ;
	{OV7670REG_COM16,	0x08},
	{OV7670REG_EDGE,	0x00},
	{OV7670REG_REG75,	0x04},
	{OV7670REG_REG76,	0x60},
	{OV7670REG_DNSTH,	0x00},
	{OV7670REG_REG77,	0x01},
	{OV7670REG_COM13,	0xc2},
	{OV7670REG_REG4B,	0x09},
	{OV7670REG_SATCTR,	0x30},
	{OV7670REG_COM16,	0x38},
	{OV7670REG_CONTRAS,	0x40},
	// ;
	{OV7670REG_ARBLM,	0x11},
	{OV7670REG_COM11,	0x02},
	{OV7670REG_NT_CTRL,	0x88},
	{0x96,				0x00},
	{0x97,				0x30},
	{0x98,				0x20},
	{0x99,				0x30},
	{0x9a,				0x84},
	{0x9b,				0x29},
	{0x9c,				0x03},
	{OV7670REG_BD50ST,	0x99},	// ;4c for 15fps
	{OV7670REG_BD60ST,	0x7F},	// ;3f for 15fps
	{0x78,				0x04},
	// ;
	{0x79,				0x01},
	{0xc8,				0xf0},
	{0x79,				0x0f},
	{0xc8,				0x00},
	{0x79,				0x10},
	{0xc8,				0x7e},
	{0x79,				0x0a},
	{0xc8,				0x80},
	{0x79,				0x0b},
	{0xc8,				0x01},
	{0x79,				0x0c},
	{0xc8,				0x0f},
	{0x79,				0x0d},
	{0xc8,				0x20},
	{0x79,				0x09},
	{0xc8,				0x80},
	{0x79,				0x02},
	{0xc8,				0xc0},
	{0x79,				0x03},
	{0xc8,				0x40},
	{0x79,				0x05},
	{0xc8,				0x30},
	{0x79,				0x26},
	// ;DS116 LENS
	{OV7670REG_LCC1,	0x20},
	{OV7670REG_LCC2,	0x20},
	{OV7670REG_LCC3,	0x07},
	{OV7670REG_LCC4,	0x00},
	{OV7670REG_LCC5,	0x05},
	{OV7670REG_LCC6,	0x07},
	{OV7670REG_LCC7,	0x0d}
};

static inline int ov7670_init(int fd)
{
	int i, j;

	// Reset OV7670 chip/device
	i = i2c_smbus_write_byte_data(fd, OV7670REG_COM7, 0x80);
	j = i2c_smbus_write_byte_data(fd, OV7670REG_COM7, 0x00);
	if ((i < 0) || (j < 0)) {
		OV7670_DBG("Reset OV7670 chip fail!\n");
		return -1;
	}

	// Set OV7670 optimal register value - from OmniVision
	for (i=0; i<sizeof(OV7670_YUV_VGA_OptimalSet)/sizeof(reginfo); i++) {
		j = i2c_smbus_write_byte_data(fd, 
			OV7670_YUV_VGA_OptimalSet[i].adr, OV7670_YUV_VGA_OptimalSet[i].val);
		if (j < 0) {
			OV7670_DBG("Set reg(%.4X) to val(%.4X) fail!\n", 
				OV7670_YUV_VGA_OptimalSet[i].adr, OV7670_YUV_VGA_OptimalSet[i].val);
			return -1;
		}
	}

	return 0;
}

int main(int argc, char **argv)
{
	int i, fd;
	long funcs;
	char fname[32];

	strncpy(fname, "/dev/i2c/0", sizeof(fname));
	fd = open(fname, O_RDWR);
	if (fd <= 0) {
		OV7670_DBG("open device (%s) fail!\n", fname);
		return -1;
	}

	// check adapter functionality
	i = ioctl(fd, I2C_FUNCS, &funcs);
	if (i < 0) {
		OV7670_DBG("ioctl I2C_FUNCS fail!\n");
		goto done;
	}
	// use FORCE so that we can write registers even when
	//	a driver is also running
	i = ioctl(fd, I2C_SLAVE_FORCE, OV7670_I2C_IOADDR);
	if (i < 0) {
		OV7670_DBG("ioctl I2C_SLAVE_FORCE fail!\n");
		goto done;
	}

	// Init CMOS - OV7670 chip/device
	i = ov7670_init(fd);
	if (i < 0) {
		goto done;
	}

done:
	close(fd);

	return 0;
}
