/*
    i2cset.c - A user-space program to write an I2C register.
    Copyright (c) 2001  Frodo Looijaard <frodol@dds.nl>, and
    Mark D. Studebaker <mdsxyz123@yahoo.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include "i2c-dev.h"

#define	USE_SENSOR_DRIVER_INTERFACE	0

#if USE_SENSOR_DRIVER_INTERFACE
#include <video/plmedia.h>
#endif

static int ov_read_byte_data(int fd, int reg)
{
#if USE_SENSOR_DRIVER_INTERFACE
	REGISTERINFO rinfo;
	rinfo.uRegister = reg;
	ioctl(fd, PLSENSOR_READ_REGISTER, &rinfo);
	return rinfo.uValue;
#else
	int data = 0;
	write(fd, &reg, 1);
	read(fd, &data, 1);
	return data;
#endif
}

static int ov_write_byte_data(int fd, int reg, int val)
{
#if USE_SENSOR_DRIVER_INTERFACE
	REGISTERINFO rinfo;
	rinfo.uRegister	= reg;
	rinfo.uValue	= val;
	ioctl(fd, PLSENSOR_WRITE_REGISTER, &rinfo);
	return 0;
#else
	return i2c_smbus_write_byte_data(fd, reg, val);
#endif
}
int main(int argc, char *argv[])
{
	int fd;
	int reg, val;
	char fname[32];

#if USE_SENSOR_DRIVER_INTERFACE
	strncpy(fname, "/dev/sensor", sizeof(fname));
#else
	strncpy(fname, "/dev/i2c/0", sizeof(fname));
#endif

	fd = open(fname, O_RDWR);
	if (fd <= 0) {
		printf("open device (%s) fail!\n", fname);
		return -1;
	}

#if !USE_SENSOR_DRIVER_INTERFACE
	{
		int i;
		long funcs;
		// check adapter functionality
		i = ioctl(fd, I2C_FUNCS, &funcs);
		if (i < 0) {
			printf("ioctl I2C_FUNCS fail!\n");
			goto done;
		}
		// use FORCE so that we can write registers even when
		//	a driver is also running
		i = ioctl(fd, I2C_SLAVE_FORCE, 0x21);
		if (i < 0) {
			printf("ioctl I2C_SLAVE_FORCE fail!\n");
			goto done;
		}
	}
#endif

	if (argc <= 1) {
		for (reg=0; reg<0xFF; reg++) {
			val = ov_read_byte_data(fd, reg);
			printf("0x%.2x 0x%.2x\n", reg, val);
		}
		printf("\n");
	} else {
		FILE *fp;
		char *p;
		char buffer[128];
//		char buffer2[128];

		fp = fopen(argv[1], "r");
		if (NULL != fp) {
			do {
				memset(buffer, 0, sizeof(buffer));
				fgets(buffer, sizeof(buffer), fp);
				if (strlen(buffer) <= 2) {
					printf("done\n");
					break;
				}
				if ((';' == buffer[0]) || ('#' == buffer[0])) {
					continue;
				}

				p = strchr(buffer, ' ');
				if ((NULL != p) && (NULL != strchr(p+1, ' '))) {
					sscanf(p+1, "%x %x", &reg, &val);
				} else {
					sscanf(buffer, "%x %x", &reg, &val);
				}
printf("reg=(0x%.2x), val=(0x%.2x)\n", reg, val);
				ov_write_byte_data(fd, reg, val);
			} while(1);
		fclose(fp);
		} else {
			printf("open (%s) fail\n", argv[1]);
		}
	}

#if !USE_SENSOR_DRIVER_INTERFACE
done:
#endif
	close(fd);

	return 0;

	ov_write_byte_data(0, 0, 0);
}
