/* iksemel (XML parser for Jabber)
** Copyright (C) 2000-2004 Gurer Ozen <madcat@e-kolay.net>
** This code is free software; you can redistribute it and/or
** modify it under the terms of GNU Lesser General Public License.
*/

#include "common.h"
#include "iksemel.h"
#include <SYS_log.h>
#include <conf_sec.h>
#include <PRO_file.h>
#define max(a,b) (a>b?a:b)
#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#endif
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <SYS_log.h>
/* stuff we keep per session */
struct session {
	iksparser *prs;
	iksid *acc;
	char *pass;
	int features;
	int authorized;
	int counter;
	int set_roster;
	int job_done;
};

/* precious roster we'll deal with */
iks *my_roster;

/* out packet filter */
iksfilter *my_filter;

/* connection time outs if nothing comes for this much seconds */
int opt_timeout = 30;

/* connection flags */
int opt_use_tls=0;
#if SC_SSL_PROXY == 1
int opt_use_sasl=0;
#else
int opt_use_sasl=0; // if server use sasl, it will be set to "1" later
#endif
int opt_log=0;

enum
{
	JL_OK,
	JL_NORMAL_ERROR,
	JL_AUTHEN_FAILED,
	JL_SERVER_DISCONNECTED,
	JL_CANNOT_CONNECT,

};
void
j_error (char *msg)
{
	fprintf (stderr, "jabber main: %s\n", msg);
	//exit (2);
}

int
on_result (struct session *sess, ikspak *pak)
{
	iks *x;

	if (sess->set_roster == 0) {
		x = iks_make_iq (IKS_TYPE_GET, IKS_NS_ROSTER);
		iks_insert_attrib (x, "id", "roster");
		iks_send (sess->prs, x);
		iks_delete (x);
	} else {
		iks_insert_attrib (my_roster, "type", "set");
		iks_send (sess->prs, my_roster);
	}
	return IKS_FILTER_EAT;
}

int
on_stream (struct session *sess, int type, iks *node)
{
	sess->counter = opt_timeout;

	switch (type) {
		case IKS_NODE_START:
			if (opt_use_tls && !iks_is_secure (sess->prs)) {
				iks_start_tls (sess->prs);
				break;
			}
			if (!opt_use_sasl) {
				iks *x;

				x = iks_make_auth (sess->acc, sess->pass, iks_find_attrib (node, "id"));
				iks_insert_attrib (x, "id", "auth");
				iks_send (sess->prs, x);
				iks_delete (x);
			}
			break;

		case IKS_NODE_NORMAL:
			if (strcmp ("stream:features", iks_name (node)) == 0) {
				sess->features = iks_stream_features (node);
				if (opt_use_sasl) {
					if (opt_use_tls && !iks_is_secure (sess->prs)) break;
					if (sess->authorized) {
						iks *t;
						if (sess->features & IKS_STREAM_BIND) {
							t = iks_make_resource_bind (sess->acc);
							iks_send (sess->prs, t);
							iks_delete (t);
						}
						if (sess->features & IKS_STREAM_SESSION) {
							t = iks_make_session ();
							iks_insert_attrib (t, "id", "auth");
							iks_send (sess->prs, t);
							iks_delete (t);
						}
					} else {
						if (sess->features & IKS_STREAM_SASL_MD5)
							iks_start_sasl (sess->prs, IKS_SASL_DIGEST_MD5, sess->acc->user, sess->pass);
						else if (sess->features & IKS_STREAM_SASL_PLAIN)
							iks_start_sasl (sess->prs, IKS_SASL_PLAIN, sess->acc->user, sess->pass);
					}
				}
			} else if (strcmp ("failure", iks_name (node)) == 0) {
				j_error ("sasl authentication failed");
				return JL_AUTHEN_FAILED;
			} else if (strcmp ("success", iks_name (node)) == 0) {
				sess->authorized = 1;
				iks_send_header (sess->prs, sess->acc->server);
			} else {
				ikspak *pak;

				pak = iks_packet (node);
				iks_filter_packet (my_filter, pak);
				if (sess->job_done == 1) return IKS_HOOK;
			}
			break;

		case IKS_NODE_STOP:
			j_error ("server disconnected");
			return JL_SERVER_DISCONNECTED;

		case IKS_NODE_ERROR:
			j_error ("stream error");
			return JL_NORMAL_ERROR;
	}

	if (node) iks_delete (node);
	return IKS_OK;
}

int
on_error (void *user_data, ikspak *pak)
{
	j_error ("authorization failed");
	return IKS_FILTER_EAT;
}

int
on_roster (struct session *sess, ikspak *pak)
{
	my_roster = pak->x;
	sess->job_done = 1;
	return IKS_FILTER_EAT;
}

void
on_log (struct session *sess, const char *data, size_t size, int is_incoming)
{
	if (iks_is_secure (sess->prs)) fprintf (stderr, "Sec");
	if (is_incoming) fprintf (stderr, "RECV"); else fprintf (stderr, "SEND");
	fprintf (stderr, "[%s]\n", data);
}

void
j_setup_filter (struct session *sess)
{
	if (my_filter) iks_filter_delete (my_filter);
	my_filter = iks_filter_new ();
	iks_filter_add_rule (my_filter, (iksFilterHook *) on_result, sess,
		IKS_RULE_TYPE, IKS_PAK_IQ,
		IKS_RULE_SUBTYPE, IKS_TYPE_RESULT,
		IKS_RULE_ID, "auth",
		IKS_RULE_DONE);
	iks_filter_add_rule (my_filter, on_error, sess,
		IKS_RULE_TYPE, IKS_PAK_IQ,
		IKS_RULE_SUBTYPE, IKS_TYPE_ERROR,
		IKS_RULE_ID, "auth",
		IKS_RULE_DONE);
	iks_filter_add_rule (my_filter, (iksFilterHook *) on_roster, sess,
		IKS_RULE_TYPE, IKS_PAK_IQ,
		IKS_RULE_SUBTYPE, IKS_TYPE_RESULT,
		IKS_RULE_ID, "roster",
		IKS_RULE_DONE);
}

int
j_connect (struct session* sess,char *jabber_id, char *pass, char* nickname)
{
	int e;

	sess->prs = iks_stream_new (IKS_NS_CLIENT, sess, (iksStreamHook *) on_stream);
	if (opt_log) iks_set_log_hook (sess->prs, (iksLogHook *) on_log);
	sess->acc = iks_id_new (iks_parser_stack (sess->prs), jabber_id);
	if (NULL == sess->acc->resource) {
		/* user gave no resource name, use the default */
		char *tmp;
		tmp = iks_malloc (strlen (sess->acc->user) + strlen (sess->acc->server) + 9 + 3);
		sprintf (tmp, "%s@%s/%s", sess->acc->user, sess->acc->server, "sercomm");
		sess->acc = iks_id_new (iks_parser_stack (sess->prs), tmp);
		iks_free (tmp);
	}
	sess->pass = pass;
	sess->set_roster = 0;
	j_setup_filter (sess);

	e = iks_connect_tcp (sess->prs, sess->acc->server, IKS_JABBER_PORT);
	switch (e) {
		case IKS_OK:
			break;
		case IKS_NET_NODNS:
			j_error ("hostname lookup failed");
			return JL_CANNOT_CONNECT;
		case IKS_NET_NOCONN:
			j_error ("connection failed");
			return JL_CANNOT_CONNECT;
		default:
			j_error ("io error");
			return JL_NORMAL_ERROR;
	}

	sess->counter = opt_timeout;
	while (1) {
		e = iks_recv (sess->prs, 1);
		if (IKS_HOOK == e) break;
		if (IKS_NET_TLSFAIL == e) 
		{
			j_error ("tls handshake failed");
			return JL_NORMAL_ERROR;
		}
		if (IKS_OK != e) 
		{
			j_error ("io error");
			return JL_NORMAL_ERROR;
		}
		sess->counter--;
		if (sess->counter == 0) 
		{
			j_error ("network timeout");
			return JL_NORMAL_ERROR;
		}
	}
	{
		iks* cc;
		cc=iks_make_pres(IKS_SHOW_AVAILABLE, nickname);
		iks_send(sess->prs, cc);
		iks_delete(cc);
	}

	return JL_OK;
}
int log_connect(char* name)
{
	int					sockfd;
	struct sockaddr_un	servaddr;

	sockfd = socket(AF_LOCAL, SOCK_DGRAM, 0);

	unlink(name);
	bzero(&servaddr, sizeof(servaddr));
	servaddr.sun_family = AF_LOCAL;
	strcpy(servaddr.sun_path, name);

	bind(sockfd, (struct sockaddr *) &servaddr, sizeof(servaddr));

	return sockfd;
}
int j_SendMessage(struct session* sess, char* jid, char* message)
{
	iks* cc;
	int ret;
	cc=iks_make_msg(IKS_TYPE_CHAT, jid, message);
	ret = iks_send(sess->prs, cc);
	iks_delete(cc);
	return ret;
}
#define ERROUT(val) {ret = val; goto errout;}
int j_main (char* server, char* account, char* password, char* sendto, char* nickname)
{
	int ret=JL_OK,log_fd=0,j_fd=0;
	struct session sess;
	int maxfd;
	fd_set rset;
//	struct LogJabberConf conf;
	char buf[1024];
	struct timeval tv;
		
	/*init some values*/
	opt_log = 0;
	memset (&sess, 0, sizeof (struct session));	
	
//	/*get values from configuration file*/
//	if(LOGReadJabberConf(&conf)!=L_OK)
//		ERROUT(JL_NORMAL_ERROR);
	
	/*setup jabber client*/
	sprintf(buf, "%s@%s", account, server);
	if ( (ret = j_connect(&sess,buf, password, nickname)) != JL_OK)
		goto errout;
	j_fd = iks_fd(sess.prs);
	
	/*setup connection with SYS_log*/
	if( (log_fd = log_connect(SYSLOG_IM_DAEMON))==0)
		ERROUT(JL_NORMAL_ERROR);

	AddLog(SYSLOG_SECTION_JABBER, 2, 4);//4=JABBER: Login successfully
	FD_ZERO(&rset);
	tv.tv_sec = 15*60;
	tv.tv_usec = 0;		
	for(;;)
	{
		FD_SET(log_fd, &rset);
		FD_SET(j_fd, &rset);
		maxfd = max(log_fd, j_fd)+1;
		select(maxfd, &rset, NULL, NULL, &tv);
		/*get messages from SYS_log and send them to Jabber server*/
		if(FD_ISSET(log_fd, &rset))
		{
			ret = read(log_fd, buf, 1024);
			if( ret <= 0)
				ERROUT(JL_SERVER_DISCONNECTED);
			buf[ret]='\0';
			if( (ret =j_SendMessage(&sess, sendto, buf)) != JL_OK)
			{
				AddLog(SYSLOG_SECTION_JABBER, 2, 6);//6=JABBER:Send message failed
				ERROUT(JL_SERVER_DISCONNECTED);
			}
			else
			{
				AddLog(SYSLOG_SECTION_JABBER, 2, 5);//5=JABBER:Send message successfully
			}
		}
		/*skip any messages from jabber server*/
		if(FD_ISSET(j_fd, &rset))
		{
			ret = read(j_fd, buf, 1024);
			if( ret <= 0)
				ERROUT(JL_NORMAL_ERROR);			
		}
		/* keep alive message*/
		if(tv.tv_sec == 0)
		{
			tv.tv_sec = 15*60;
			tv.tv_usec = 0;		
			buf[0]='\x20';
			write(j_fd, buf, 1);
		}
	}
errout:
	if(log_fd)
		close(log_fd);
	if(j_fd)
		close(j_fd);
	if(sess.prs)
		iks_parser_delete (sess.prs);	
	if( ret == JL_CANNOT_CONNECT || JL_NORMAL_ERROR == ret)
		AddLog(SYSLOG_SECTION_JABBER, 2, 2);//2=JABBER: Can't connect to server
	if(ret == JL_AUTHEN_FAILED)
		AddLog(SYSLOG_SECTION_JABBER, 2, 1);//1=JABBER: Authentication failed
	if(ret == JL_SERVER_DISCONNECTED)
		AddLog(SYSLOG_SECTION_JABBER, 2|SYSLOG_FILEONLY, 1);//3=JABBER: Server disconnected

	return ret;
}
int create_stunnel(char* server)
{
	int fd;
	char buf[128];	
	char fix[]="foreground = no\ndebug=0\npid =\noptions = ALL\nTIMEOUTclose = 0\nclient = yes\nverify = 0\n[jabber]\naccept = 5222\n";
	system("/usr/bin/killall jabber_stunnel");	
	//make jabber_stunnel.conf
	fd = open("/tmp/jabber_stunnel.conf", O_RDWR | O_CREAT | O_TRUNC);
	if(fd < 0)
		return -1;
	write(fd, fix, sizeof(fix)-1);
	sprintf(buf, "connect = %s:5223\n", server);
	write(fd, buf, strlen(buf));
	close(fd);
	system("/usr/local/bin/jabber_stunnel /tmp/jabber_stunnel.conf >/dev/zero 2>/dev/zero");
	return 0;	
}
int main (int argc, char *argv[])
{
	FILE *fp;
	char host_name[32];
	struct LogJabberConf conf;
	int ret;
	
	LOGReadJabberConf(&conf);
	if(conf.enable == 0)
		return 0;
	fp = fopen(SYSTEM_CONF,"rt");
	ret = PRO_GetStr(SEC_SYS,SYS_HOSTNAME,host_name, 31, fp);

	if(create_stunnel(conf.server))
		return -1;
	
	while(1)
	{
		j_main(conf.server, conf.account, conf.password, conf.sendto, host_name);
		sleep(1*60);	//we should reconnect to jabber server after 1min when error occured,15min is too long for consumer to wait
	}
	system("/usr/bin/killall jabber_stunnel >/dev/null 2>/dev/null ");
	return 0;
}

